
#ifndef FIXEDLENTS_HH
#define FIXEDLENTS_HH

#include "TSeries.hh"
#include <stdexcept>

/**  The fixed length time series is defined in
  *
  *  \begin{verbatim}
     #include "FixedLenTS.hh"
     \end{verbatim}
  *  You set the maximum time either in the constructor, e.g.
  *  \begin{verbatim}
     FixedLenTS mHistory(7200.0);
     \end{verbatim}
  *  Or explicitly 
     \begin{verbatim}
     FixedLenTS mHistory;
     mHistory.setMaxLen(t);
     \end{verbatim}
  *  You then append data like....
  *  \begin{verbatim}
     //----------------------------------  Keep the history
     Time t0 = fs.getStartTime();
     Interval dT = fs.getEndTime()-t0;
     mTimeTot += dT;
     float bandP = mBand;
     mHistory.fixedAppend(t0, dT, &bandP);
     \end{verbatim}
  *  When it reaches the maximum it starts to delete data from the start,
  *  rather than increasing the length. The FixedLenTS is based on a TSeries 
  *  so you can reference it as you would a TSeries.
  *  @memo Fixed length Time series
  *  @author J. Zweizig
  *  @version 1.2; Last modified February 17, 2006
  */
class FixedLenTS : public TSeries {
public:

  /**  Default constructor creates a fixed length TSeries with zero length.
    *  A #FixedLenTS# created with this constructor is unusable  until the 
    *  maximum length has been set with #setMaxLen()#
    *  @memo Default constructor.
    */
  FixedLenTS(void);

  /**  Construct a #FixedLengthTS# with a specified maximum length.
    *  @memo Data constructor.
    *  @param max maximum length for time series.
    */ 
  explicit FixedLenTS(Interval max);

  /**  Create a FixedLenTS that is an identical copy of the argument series.
    *  @memo Copy constructor.
    */
  FixedLenTS(const FixedLenTS& x);

  /**  Destroy the Fixed length time series.
    *  @memo Destructor.
    */
  inline virtual ~FixedLenTS() {}

  /**  Append one or more float data words to the series. If the series 
    *  with the appended data is longer than the specified maximum length,
    *  data are removed from the beginning of the series and the start
    *  time is updated as appropriate.
    *  @memo Append float data.
    *  @param t0 Time of first datum.
    *  @param dT Time step between data.
    *  @param x  Pointer to one or more data.
    *  @param n  Number of data to be appended to the time series.
    */
  void fixedAppend(const Time& t0, Interval dT, float* x, int n=1);

  /**  Append a TSeries to the fixed length series. If the resulting series 
    *  would be longer than the specified maximum length, data are removed 
    *  from the start of the current series and/or the start of the data to 
    *  be appended and the current series start time is updated as appropriate.
    *  @memo Append a TSeries.
    *  @param ts Teries to be appended.
    */
  void fixedAppend(const TSeries& ts);

  /**  Append zeros to the series until the specified end-time.
    *  @memo Append zeroes until the specified time.
    *  @param t0 Time of first datum.
    *  @param dT Time step between data.
    */
  void padUntil(const Time& t0, Interval dT);

  /**  Set the maximum length for the fixed length time series.
    *  @memo specify the maximum length.
    *  @param dt Maximum length
    */
  void setMaxLen(Interval dt);

  /**  Get the Maximum length specified.
    *  @memo Get the Maximum length.
    *  @return aximum length.
    */
  Interval getMaxLen(void) const;

private:
  Interval maxLen;
};

//======================================  Inline functions
inline
FixedLenTS::FixedLenTS(void)
  : maxLen(0) 
{}

inline
FixedLenTS::FixedLenTS(Interval max)
  : maxLen(max) 
{}

inline
FixedLenTS::FixedLenTS(const FixedLenTS& x)
  : TSeries(x), maxLen(x.maxLen)
{}

inline void 
FixedLenTS::setMaxLen(Interval dt){
    maxLen = dt;
}

inline Interval
FixedLenTS::getMaxLen(void) const {
    return maxLen;
}

inline void
FixedLenTS::fixedAppend(const TSeries& ts) {
    if (ts.isEmpty()) return;
    Time tStart = ts.getEndTime() - maxLen;
    if (!isEmpty() && getEndTime() <= tStart) {
        if (tStart > getStartTime()) eraseStart(getStartTime()-tStart);
	if (getEndTime() < ts.getStartTime()) extend(ts.getStartTime());
	TSeries::Append(ts);
    } else if (ts.getStartTime() < tStart) {
        TSeries::operator=(ts.extract(tStart, maxLen));
    } else {
        TSeries::operator=(ts);
    }
}

inline void
FixedLenTS::fixedAppend(const Time& t0, Interval dT, float* x, int n) {
    if (!dT) throw std::runtime_error("FixedLenTS::fixedAppend invalid dT");
    if (isEmpty()) {
        int nBins = int(maxLen / dT + 0.5);
	if (!refDVect()) TSeries::operator=(TSeries(t0, dT, nBins));
        setData(t0, dT, x, n);
    } else {
        if (dT != getTStep()) 
	    throw std::runtime_error("FixedLenTS::fixedAppend invalid dT");
        Interval dTot = t0 + double(n)*dT - getStartTime();
        if (dTot > maxLen) eraseStart(dTot - maxLen);
        extend(t0);
	Append(t0, dT, x, n);
    }
}

inline void
FixedLenTS::padUntil(const Time& t0, Interval dT) {
    if (!dT) throw std::runtime_error("FixedLenTS::padUntil invalid dT");
    Time tStart = t0 - maxLen;
    if (isEmpty() || getEndTime() <= tStart) {
	float data(0.0);
        setData(tStart, dT, &data, 1);
    } else if (tStart > getStartTime()) {
	eraseStart(tStart - getStartTime());
    }
    if (dT != getTStep()) 
        throw std::runtime_error("FixedLenTS::padUntil invalid dT");
    extend(t0);
}

#endif // FIXEDLENTS
