#ifndef DAQ_ERR_FILTER_HH
#define DAQ_ERR_FILTER_HH

#include "Pipe.hh"

/**  This class simulates DAQ errors. At present the error types included 
  *  are:
  *  \begin{itemize}
  *  \item {\bf Word swapping:} during S1 several 16-bit word variables had
  *       the order each sample pair swapped. This mode operates only on 
  *       2-byte data time series.
  *  \item {\bf LSB errors:} errors occurred randomly in the Least Significant
  *       Byte (LSB) of 16-byte data transfers during science runs S1 - S3. 
  *       The simulation code generates errors
  *       by jamming #0xff# into the low order byte of random words. Fixed 
  *       or random width clusters are generated at random intervals as
  *       specified by the #setClusterWidth#, #setAvgWidth# and #setErrRate# 
  *       methods. The cluster start time is calculated by generating a
  *       random time offset from an exponential distribution with average
  *       time as specified by #setRate#. The time is adjusted to the 
  *       previous 1/16th second boundary. If an average cluster width was 
  *       specified, a random cluster width is generated from a Poisson
  *       distribution with the specified average. If the average with
  *       is zero, the specified constant cluster width is used. If a width
  *       specified or generated to be zero, no cluster is generated at the
  *       calculated start time.
  *  \end{itemize}
  *  The types of errors to be simulated are set by either the constructor
  *  argument or by the setSubTypes method. The type string is a comma 
  *  separated list of error types with one or more of "swap" or "lsb".
  *  @author J. Zweizig
  *  @version 1.0; Last modified: May 5, 2004
  */
class DaqErrFilter : public Pipe {
public:
  enum type_enum {
    tSwap,
    tLSB,
    tMax
  };

public:
  using Pipe::apply;
  using Pipe::dataCheck;

  /**  Daq error filter constructor. The argument is a list of error types 
    *  to be simulated with syntax as specified in the setSubTypes method.
    *  @memo Daq error constructor.
    *  @param subtype List of errors to be simulated.
    */
  DaqErrFilter(const char* subtype);

  /**  Daq error destructor.
    */
  ~DaqErrFilter(void);

  //------------------------------------  Set parameters
  /** Return a string containing a list of error tyeps to be simulated.
   */
  std::string getTypeString(void) const;

  /** Enable simulation of the specified error type.
   */
  bool selectType(const std::string& s);

  /**  Specify an average number of errors to be generated in a cluster of 
    *  LSB errors. By default the average is ?? 
    */
  void setAvgWidth(double avgWid);

  /**  Specify a fixed number of errors in a cluster. Note that to specify a 
    *  fixed cluster widthm the average clister width must be zero.
    */
  void setClusterWidth(long wid);

  void setErrRate(double F);

  /**  Set the error tyopes to be simulated by this filter. The error types
    *  are specified as a comma separated list of error names. At present 
    *  the valid type names are "swap" and "lsb".
    */
  void setSubTypes(const std::string& s);

  //--------------------------------  Pipe methods
  TSeries apply(const TSeries& x);
  DaqErrFilter* clone(void) const;
  void dataCheck(const TSeries& x) const;
  Time getStartTime(void) const;
  Time getCurrentTime(void) const;
  bool inUse(void) const;
  void reset(void);

private:
  void roll(void);
  bool testType(type_enum t) const;

private:
  long   mSubTypes;
  long   mClustWidth;
  double mAvgWidth;
  double mErrRate;
  Time   mLsbErrTime;
  Time   mStartTime;
  Time   mCurrentTime;
};

//======================================  Inline methods
inline void 
DaqErrFilter::dataCheck(const TSeries& x) const {
}

inline Time
DaqErrFilter::getCurrentTime() const {
    return mCurrentTime;
}

inline Time
DaqErrFilter::getStartTime() const {
    return mStartTime;
}

inline bool
DaqErrFilter::inUse(void) const {
    return mStartTime != Time(0);
}

inline bool 
DaqErrFilter::testType(type_enum t) const {
    return ((1<<t) & mSubTypes) != 0;
}

#endif // !defined(DAQ_ERR_FILTER_HH)
