#ifndef LAUDIO_HH
#define LAUDIO_HH

//
//  Starting implementation will assume the solaris devaudio API. For the
//  future I think that the OSS API looks promising.
#ifdef sun
#define SOLARIS_AUDIO
#else
#define OPEN_SOUND_SYSTEM
#endif

#ifndef __CINT__
#include <string>
#include <stdexcept>
#include <iosfwd>
#else
namespace std {
  class runtime_error;
}
#endif

#ifdef SOLARIS_AUDIO
struct audio_info;
struct audio_prinfo;
#endif

/**  LAudio class is an API for audio input and output. The Version 2.0
  *  implementation uses the Solaris audio driver on sun machines and 
  *  the Open Sound System (OSS) driver on Linux and other systems.
  *  @memo Audio I/O API
  *  @author John Zweizig
  *  @version 2.0; Modified May 22, 2001.
  */
class LAudio {
public:

  /**  Data type used internally for data counting.
    */
  typedef unsigned int count_t;

  /**  The mode specified whether the audio interface will beused for
    *  listening (Input from a microphone), Singing (output to a speaker),
    *  monitoring (Input from a microphone with an immediate copy going to 
    *  the speaker) or Conversation (Listening and talking).
    */
  enum iomode {m_listen, m_sing, m_monitor, m_converse};

  enum ioport {p_Speaker=1, p_Phone=2, p_Line=4};

  /**  Create an interface to the default audio device. The default 
    *  device is usually /dev/audio, but this default may be overridden
    *  by setting the DEVAUDIO environment vaariable.
    *  @memo Connect to the default device.
    */
  LAudio(void);

  /**  Create an interface and connect to the spcified device.
   */
  LAudio(const char* device);

  /**  Disconnect/destroy the audio interface.
    *  @memo destroy the interface.
    */
  ~LAudio(void);

  /**  Close a previously opened audio device. The close() request hangs 
    *  while data are drained from the output buffer.
    *  @memo close the audio device.
    */
  void close(void);

  /**  Block further execution until all samples have been sent to the 
    *  the audio device.
    *  @memo Wait for data to be played.
    */
  void drain(void) throw(std::runtime_error);

  /**  Print out the status of the interfaces in use.
    *  @memo Dump interface status.
    */
  void dumpStatus(std::ostream& out) const;

  /**  The device type is read from the nterface card. It is returned as
    *  an ASCII string in the form <interface>-<version>-<configuation>.
    *  @memo Get the device type.
    */
  std::string getDevice(void) const;

  /**  The signal may be encoded with either a "Linear" transformation, 
    *  the mu-transformation ("mu-Law") or the A-transformation ("A-Law").
    *  @memo Get the signal encoding.
    */
  const char* getEncoding(void) const throw(std::runtime_error);

  /**  Set the gain for feeding data from the microphone to the speaker
    *  when in monitor mode.
    */
  count_t getMonitorGain(void) const throw(std::runtime_error);

  /**  Get the data bit precision. 8-bit and 16-bit data words are 
    *  supported.
    *  @memo Get data precision.
    */
  count_t getPrecision(void) const throw(std::runtime_error);

  /**  Get the Sample rate.
    *  @memo Get data sample rate.
    *  @return Data sample rate in Hz.
    */
  count_t getSpeed(void) const throw(std::runtime_error);

  /**  Get the output amplifier gain for feeding the speaker
    *  when in monitor sing, monitor or converse mode.
    */
  double getVolume(void) const throw(std::runtime_error);

  /**  Test whether the interface has a device connected to it.
    *  @memo test if device is connected.
    */
  bool isOpen(void) const;

  /**  Enable/Disable all sound outputs. All volume controls and output
    *  port selections are left intact.
    *  @memo Mute audio outputs.
    *  @param mute True if audio output is to be muted (disabled).
    */
  void mute(bool mute) throw(std::runtime_error);

  /**  Open the specified output device. If a NULL pointer or zero-length
    *  device name is specified, the device specified by the #AUDIODEV# 
    *  environment variable is used or if #AUDIODEV# is not specified,
    *  "/dev/audio" is used.
    *  @memo  Open an audio device.
    *  @param device pointer to the device name string.
    *  @param mode   Input/Output mode.
    */
  void open(const char* device=0, iomode mode=m_sing);

  /**  Interpolate samples where necessary to match the sampling frequency
    *  to the audio output rate and send the result to the audio output port.
    *  @memo Blast out an interpolated wave form.
    *  @param freq   Sample frequency (in Hz) of input data.
    *  @param data   A short integer array containing the samples to be sung.
    *  @param length Number of input data samples.
    */
  void reSample(double freq, short data[], count_t length);

  /**  Interpolate samples where necessary to match the sampling frequency
    *  to the audio output rate and send the result to the audio output port.
    *  @memo Blast out an interpolated wave form.
    *  @param freq   Sample frequency (in Hz) of input data.
    *  @param data   A float array containing the samples to be sung.
    *  @param length Number of input data samples.
    */
  void reSample(double freq, float data[], count_t length);

  /**  Select the specified encoding mode. Valid encoding modes are: 
    *  #Linear#, #A-Law# and #u-Law#. At present only #Linear# is 
    *  implemented. A runtime error is thrown if the specified encoding 
    *  technique is not known or if the the selection is rejected by the 
    *  device driver.
    *  @memo Select data encoding technique.
    *  @param encode Pointer to a string specifiying the encoding mode.
    */
  void setEncoding(const char* encode) throw(std::runtime_error);

  /**  Set a linear transformation for input data.
    */
  void setInputTransform(double bias, double scale);

  /**  Select audio output ports. The available port IDs are described in 
    *  the ioport enumerator.
    *  @memo Select audio output ports.
    *  @param mask Bit mask with a bit set for each selected port.
    */
  void setPorts(count_t mask) throw(std::runtime_error);

  /**  Set the monitor gain. The monitor gain controls the amplitude 
    *  with which the input port data are copied to the output port(s).
    *  @memo Set monitor gain.
    *  @param gain Multiplier for input to be added to the output.
    */
  void setMonitorGain(count_t gain) throw(std::runtime_error);

  /**  Set the sample rate. Because of hardware limitations in some sound
    *  cards, the should be set only after the data format and number of 
    *  channels have been set.
    *  @memo rate Sample rate to be used.
    */
  void setSpeed(count_t rate) throw(std::runtime_error);

  /**  Set number of bits of precision for each audio sample. 8-bit and 
    *  16-bit precision are currently allowed. Setting the precision 
    *  implies a Linear encoding format.
    *  @memo bits Length of each sample in bits.
    */
  void setPrecision(count_t bits) throw(std::runtime_error);

  /**  Set the amplifier gain for feeding the speaker.
    */
  void setVolume(double volume) throw(std::runtime_error);

  /**  Spew audio output.
    *  @memo Blast out a sampled wave form.
    *  @param data   A short integer array containing the samples to be sung.
    *  @param length Number of data samples.
    */
  void sing(short data[], count_t length);

private:
  int Encode(const std::string& x) const throw(std::runtime_error);
  const char* Encode(int x) const;

#ifdef SOLARIS_AUDIO
  void dumpChannel(std::ostream& out, const char* title, 
		   const audio_prinfo& info) const throw(std::runtime_error);
  void getStatus(audio_info* info) const throw(std::runtime_error);
  void setStatus(const audio_info* info) throw(std::runtime_error);
#endif
  short transform(double) const;

private:
  struct DspStatus {
      DspStatus(void);
      bool valid;
      int sample_rate;    //  Sample rate.
      int format;         //  Data encoding.
      int precision;      //  Data precision (bits)
      int buffer_size;    //  Buffer size in bytes
      double gain;        //  Gain
      int samples;        //  Number of samples transferred
  };

private:
  int    mFD;
  iomode mMode;
  bool   mBigEnd;
  DspStatus mInStat;
  DspStatus mOutStat;
  double mOutputSPS;
  int    mOutputVal;

  /**  Amount of time you must wait before spitting out another sample. 
    *  This variable is set by the resampling algorithm when the next
    *  expected sample doesn't correspond toin time to the next dac clock
    *  time.
    */
  double mDt;

  /** Input transformation to scale input values
   */
  double mBias;
  double mScale;
};

inline short
LAudio::transform(double x) const {
     return short((x-mBias)*mScale);
}

#endif  // LAUDIO_HH
