//
//    CorrMon produces a measure of interchannel correlation
//    between a signal channel and other `environmental' channels and
//    optionally uses this information to `clean' the signal.
//
//    "CorrMon -?" or "CorrMon -help" for full usage instructions 
//    Basic usage:  "CorrMon -conf config_file"
//    where config_file has the following structure:
//
//    # Comments (start with a #)
//    # More comments
//    ...  
//    # Last comment
//    Line 1: <descriptor>
//    Line 2: <correlation width>
//    Line 3: <sampling time>
//    Line 4: <window: 0=no, 1=yes>
//    Line 5: <no of sample times to average>
//    Line 6: <rho2 trigger threshold>
//    Line 7: <lower frequency monitored> 
//    Line 8: <Upper frequency monitored>
//    Line 9: <should we clean: 0=no, 1=yes>  
//    Line 10: <number of channels>
//    Line 11: <signal channel> <frequency> 
//    Line 12: <first environmental channel> <frequency>
//    ...
//    Line 10+number of channels: <last environmental channel> <frequency>

//
//    The fields above have the following meanings:
//
//    <descriptor>   A character string describing the run, used  for
//                      naming intermediate directories and files and the
//                      root canvases
//    <number of channels>   Number of channels including the signal 
//    <sample time>  The period in seconds over which correlations are sought.
//    <signal/environmental channel>      The name of the signal/environmental 
//                      channel between which correlations are sought.
//    <frequency>    The sample rate of the channel
//    <clean>        Set to 1 if you want a best estimate of the `true' spectrum
//                      based on a knowledge of correlations gleaned, 0 otherwise.
//

#include "CorrMon.hh"
#include "CorrMonUtility.hh"

#include <cstdlib>
#include <cmath>
#include <unistd.h> 
#include <stdio.h>
#include <iostream>
#include <iomanip>
//#include <strstream> 
#include <fstream>
#include "TrigRslt.hh"

//-->  The next three lines are needed if you are going to generate triggers.
//     The descriptive title in PIDTITLE should the monitor function.
#define PIDCVSHDR "$Header: /ldcg_server/common/repository_gds/gds/Monitors/CorrMon/CorrMon.cc,v 1.7 2003/04/11 21:45:14 jzweizig Exp $"
#define PIDTITLE  "Correlation monitor"
#include "ProcIdent.hh"


const int N_ROOT_COLORS=18;
const int root_colors[N_ROOT_COLORS]={2,3,4,5,6,7,8,9,11,25,28,30,35,38,41,43,46,50}; 

#ifdef CORRMON_ROOT
EXEROOT(CorrMon);
#else
EXECDAT(CorrMon); // if no root graphics
#endif

using namespace std;

CorrMon::CorrMon(int argc, const char* argv[]) 
  :DatEnv(argc, argv),MonServer("CorrMon"),mEps(false),mDebug(false),
   mLog(false),mSummary(false),mTrigger(true),mTrendFlag(false),mPasses(-1),
   highLevelTrigger(0.99),nTSBuffer(8192),nCorrBuffer(1024)
{
  // Parse command line
  const char* pconfig=0;
  for (int i=1 ; i<argc ; i++) {
    std::string argi = argv[i];
    if (argi == "-conf" || argi == "-c" || argi == "-config") {
      pconfig = argv[++i];
    } else if (argi == "-eps") {
      if (argv[0] == "CorrMonRoot") {
	mEps = true;
      } else {
	cerr << "To produce eps use CorrMonRoot (the non-interactive version"
	     << " of CorrMon) with the same arguments as CorrMon" << endl;
	exit(5);
	return;
      }    
    } else if (argi == "-debug" || argi == "-d" || argi == "-verbose") {
      mDebug = true;
    } else if (argi == "-log") {
      mLog = true;
    } else if (argi == "-summary") {
      mSummary = true;  
      summaryFileExists=false; 
    } else if (argi == "-notrig" || argi == "-notriggers") {
      mTrigger = false;
    } else if (argi == "-notrend") {
      mTrendFlag = false;
    } else if (argi == "-once" || argi == "-onepass") {
      mPasses=1;
    }  else if (argi == "-npasses" || argi == "-npass") {
      mPasses = atoi(argv[++i]);
    }
    else if (argi == "-help" || argi == "-?") {
      usage();
      exit(3);
      return;
    }  
    else {
      cerr <<  "Invalid argument: " << argi << endl;
	    usage();
	    exit(4);
	    return;    
    }
  }

  if (mDebug && mEps) std::cerr << "# CorrMon: postscript output requested" << std::endl;
  if (mDebug && mLog) std::cerr << "# CorrMon: log files requested" << std::endl;
  if (mDebug && mSummary) std::cerr << "# CorrMon: summary files requested" << std::endl;
  if (mDebug && mPasses != -1) std::cerr << "# CorrMon: " <<  mPasses << " pass(es) requested" << std::endl;

  //---------------------------------  Check file name
  if (!pconfig || !*pconfig) {
    std::cerr<< "CorrMon: configuration file not specified" << std::endl;
    exit(2);
    return;
  }

  readConfigFile(pconfig);

  getDacc().setBuffer(0);    // Take up data as requested
  getDacc().setDebug(0);
  getDacc().setStride(dTime);


  if (mDebug) std::cerr << "# CorrMon: Starting from file " << getDacc().getFile() << std::endl;

  deltaF=chanv[0].mRate/chanv[0].mPoint;

  int length=chanv[0].mPoint;
  for(int id=1;id<nChan;id++) {
    if (chanv[id].mPoint<length) length=chanv[id].mPoint;
  }
  length/=2;   // because we are dealing with the *complex* ffts

  if ( (length%correlationWidth) != 0 ) {
    length-=length%correlationWidth;
  }
  if (mDebug) std::cerr << "# CorrMon: Using " << length << 
		" FFT values (correlationWidth = " << correlationWidth << ")" << std::endl;

  if (mDebug && mWindow) std::cerr << "# CorrMon: Using a Hanning window" << std::endl;


  //***************************************************************//
  //                         Allocate memory                       //
  //***************************************************************//

  nEnvChan=nChan-1;   // because the configuration file includes the signal channel
  mA=new fComplex[nEnvChan*nEnvChan]; 
  mB=new fComplex[nEnvChan];

  rhoSq=new float[nEnvChan];
  typedef float* float_ptr;
  rhoSqPairwise=new float_ptr[nEnvChan];
  avgRhoSqPairwise=new float_ptr[nEnvChan];     
  maxRhoSqPairwise=new float_ptr[nEnvChan];     



  typedef unsigned int* uint_ptr;
  monitorCount=new uint_ptr[nEnvChan];
  triggerCount=new uint_ptr[nEnvChan];

  nRho=length/correlationWidth;
  if (nCorrBuffer > nRho ) nCorrBuffer=nRho;

  frequencies=vector<freqEntry>(nRho);
  for  (int iOffset=0;iOffset<nRho;iOffset++) {
    frequencies[iOffset].freq=(iOffset+0.5)*correlationWidth*deltaF; 
    sprintf(frequencies[iOffset].freqString,"%9.2f",frequencies[iOffset].freq);
  }

  for (int id=0;id<nEnvChan;id++) {
    rhoSqPairwise[id]=new float[nRho];
    avgRhoSqPairwise[id]=new float[nRho];
    maxRhoSqPairwise[id]=new float[nRho];
    monitorCount[id]=new unsigned int[nRho];
    triggerCount[id]=new unsigned int[nRho];
  }


  

  if (mTrendFlag) {
      cerr << " ACO : trends to be implemented " << endl;
     
//       const char *dmttrendout= getenv("DMTTRENDOUT");
//       if (dmttrendout) trendDir = string(dmttrendout)+"/";
//       if (getDebug()) cerr << "MTLineMon: Trend directory " << trendDir << endl;

//       const char *dmtoutput= getenv("DMTOUTPUT");
//       if (dmtoutput) outputDir = string(dmtoutput)+"/";
//       if (getDebug()) cerr << "MTLineMon: Index directory " << outputDir << endl;

    //   mTrend.setName(serverName.c_str());  
//       mTrend.setType(Trend::kMinute);
//       mTrend.setFrameCount(1);   
//       if (chanv.mName && chanv.mName[2] == ':') {
//         string ifo(string(chanv.mName), 2);
//         mTrend.setIFO(ifo.c_str());
//       }       

  }

  mCount=1;

 
  // get OUTPUT directories
  const char *dmthtmlout= getenv("DMTHTMLOUT");
  if (dmthtmlout) htmlDir = string(dmthtmlout)+"/";
  if (mDebug) cerr << "# CorrMon: HTML directory " << htmlDir << endl;
  // set html file name
  htmlFile=htmlDir+"Corr_"+string(chanv[0].mName)+".html";
  if (mDebug) cerr <<"# CorrMon: output html file name "<< htmlFile << endl;


  //----------------------------------  Define and initialize parameters.
   mDict.addPar("HTMLFile",   htmlFile);
   mDict.addPar("Stride",     dTime);
   mDict.addPar("MonName",    string(argv[0]));


  // set server Name
   
   string serverName = mDict.getString("MonName")+string(chanv[0].mName);
   setServerName(serverName.c_str());



  char title[256];
  sprintf(title,"%s: Original Signal",chanv[0].mName);
  serveData(title, plotTS);
  sprintf(title,"%s: Original Spectrum",chanv[0].mName);
  serveData(title,plotSpec);



  if (mClean) {
    sprintf(title,"%s: Cleaned Signal",chanv[0].mName);
    serveData(title,plotCleanedTS);
    sprintf(title,"%s: Cleaned Spectrum",chanv[0].mName);
    serveData(title,plotCleanedSpec); 
  }
}

void CorrMon::ProcessData() {

  for (int id=0;id<nChan;id++) {
    // check frames are consistent with configuration file 
    if (mDebug && chanv[id].mRate*double(chanv[id].mTS->getTStep()) != 1.0) {
      std::cerr << "# CorrMon: Sample rate for " << chanv[id].mName <<  
	" is " << 1.0/double(chanv[id].mTS->getTStep()) << " not " <<
	chanv[id].mRate << " as specified in " << config << std::endl;
      abort();
    }
   
    // remove the mean
    *(chanv[id].mTS)-=chanv[id].mTS->getAverage();

    // window the data 
    TSeries *ts=chanv[id].mTS;
    if (mWindow) {
      Hanning Wn;
      chanv[id].mFS->setData(Wn(*ts));
    }
    else {
      chanv[id].mFS->setData(*ts);
    }
    chanv[id].mFS->setName(chanv[id].mName);
  }
  
  *plotTS=TSeries(chanv[0].mTS->decimate(mPlotDecimation));
  *plotSpec=FSpectrum(*plotTS);
 
  avgRhoSqPairwiseSpec=vector<FSpectrum*>(nChan);
  for (int id=0;id<nEnvChan;id++) {      
    avgRhoSqPairwiseSpec[id]=new FSpectrum(0,correlationWidth*deltaF,chanv[0].mTS->getStartTime(),
					   dTime,nRho,avgRhoSqPairwise[id]);
    char title[256];
    sprintf(title,"Correlations between %s and %s",chanv[0].mName,chanv[id+1].mName);
    serveData(title, avgRhoSqPairwiseSpec[id],0);
  }
  
  if (mCount == 1) {
    startGPS=chanv[0].mTS->getStartTime().getS();
    TimeStr(chanv[0].mTS->getStartTime(), startTime, "%s");
    startTimeString=std::string(descriptor)+"_"+std::string(startTime);
    liveGPS=0;

    for (int id=0;id<nEnvChan;id++) {
      for  (int iOffset=0;iOffset<nRho;iOffset++) {
	avgRhoSqPairwise[id][iOffset]=0;
	maxRhoSqPairwise[id][iOffset]=0;
	monitorCount[id][iOffset]=0;
	triggerCount[id][iOffset]=0;
      }
    }

    if (mLog) {
      logFileName=startTimeString+".log";
      if (mDebug) std::cerr << "# CorrMon: Log File name " << logFileName << endl;
    }

    if (mSummary && !summaryFileExists ) {
      std::string summaryFileName=startTimeString+".summary";
      if (mDebug) std::cerr << "# CorrMon: Summary File name " << summaryFileName << endl;
      summaryFile.open(summaryFileName.c_str());
      summaryFileExists=true;
    }

  }
  
  currentGPS=chanv[0].mTS->getStartTime().getS();
  liveGPS+=Time::ulong_t(dTime);


  //***************************************************************//
  //         Step through the range of offsets and call the        // 
  //       function calc_rho where the major calculation is done   //
  //***************************************************************//

  int offset;
  FSeries* band = new FSeries[nChan];
  threshold=((0.1) < (5.0/correlationWidth) ? (0.1) : (5.0/correlationWidth) ); // see discussion in Hua et al. 

  for (int iOffset=0;iOffset<nRho;iOffset++) { 
    offset=iOffset*correlationWidth;
    for (int id=0;id<nChan;id++) { 
      band[id]=(chanv[id].mFS)->extract(offset*deltaF,correlationWidth*deltaF);
    }

    mModX2Sum=(band[0]*band[0]).Real();
    for (int col=0;col<nEnvChan;col++) {
      mB[col]=band[col+1]*band[0];
      for (int row=0;row<=col;row++) {
	mA[RC(row,col,nEnvChan)]=band[row+1]*band[col+1];
      }
    } 

    corrThreshold();
    
    // Reduction in variance using the best R for just the single channel (after thresholding)  
    if (mModX2Sum !=0 ) {
      for (int id=0;id<nEnvChan;id++) {
	rhoSqPairwise[id][iOffset]=mB[id].MagSq()/(mA[RC(id,id,nEnvChan)].Real()*mModX2Sum);
      }
    }
    
    if (mClean) {
      int info = chanClean(offset,band);
      if (info != 0) {
	cerr << "# CorrMon: chesv_ returned nonzero info = " << info << endl;
	finish();
      }
    }


  } // end of iOffset loop
  



  if (mClean) {
    *cleanedSignal.mTS=TSeries(*(chanv[0].mTS));
    *cleanedSignal.mFS=FSeries(*(chanv[0].mFS));
    cleanedSignal.mFS->setData(chanv[0].mPoint,cleanedFSData);
    cleanedSignal.mSpec->setData(*(cleanedSignal.mFS));
    *cleanedSignal.mTS=TSeries(*(cleanedSignal.mFS));
    *plotCleanedTS=TSeries(cleanedSignal.mTS->decimate(mPlotDecimation));
    *plotCleanedSpec=FSpectrum(*(cleanedSignal.mFS));
  }

  if (mCount <= nAveragingTimes) {
    for (int id=0;id<nEnvChan;id++) {
      for  (int iOffset=0;iOffset<nRho;iOffset++) {
	avgRhoSqPairwise[id][iOffset]*=(mCount-1);
	avgRhoSqPairwise[id][iOffset]+=rhoSqPairwise[id][iOffset];
	avgRhoSqPairwise[id][iOffset]/=mCount;
        if (rhoSqPairwise[id][iOffset] > maxRhoSqPairwise[id][iOffset]) 
	  maxRhoSqPairwise[id][iOffset]=rhoSqPairwise[id][iOffset];
	if ( rhoSqPairwise[id][iOffset] > lowLevelTrigger ) {
	  monitorCount[id][iOffset]++;
	  if (mDebug) std::cerr << "# CorrMon: Low level trigger threshold exceeded for channel " << chanv[id+1].mName 
			      << " at frequency " << frequencies[iOffset].freq << "Hz (band: "
			      << iOffset  << ")" << std::endl;
	}
	if (mTrigger &&  rhoSqPairwise[id][iOffset] > highLevelTrigger ) {
	  rhoSqTrigger(id,iOffset);
	  triggerCount[id][iOffset]++;
	} 
      }
    }
  }


  bool emptyLog=true;  
  if (mCount == nAveragingTimes) { 
   
    for (int id=0;id<nEnvChan;id++) {
      for (int iOffset=0;iOffset<nRho;iOffset++) {

	if (monitorCount[id][iOffset]) {
	  //***************************************************************//
	  //    Create a log file if this is the first trigger             //
	  //     in this section of data                                   //
	  //***************************************************************//
	  if (emptyLog) {
	    logFile.open(logFileName.c_str());
	    emptyLog=false;
	  }

	  if (mLog) logFile << chanv[id+1].mName 
			    << " at frequency " << (iOffset+0.5)*correlationWidth*deltaF << "Hz (band: "
			    << iOffset  << ") : mean rho2 = " <<  avgRhoSqPairwise[id][iOffset]  
                            << "    max rho2 = " <<  maxRhoSqPairwise[id][iOffset] <<   endl;

	  if (mSummary) {
	    summaryFile << setw(10) << startTime << "\t"; 
	    summaryFile << setw(20) << chanv[id+1].mName << "\t"; 
	    summaryFile << setw(10) << (iOffset+0.5)*correlationWidth*deltaF << "\t";
	    summaryFile << iOffset  << "\t"; 
	    summaryFile << avgRhoSqPairwise[id][iOffset] <<  "\t"; 
	    summaryFile << maxRhoSqPairwise[id][iOffset] <<  endl;
	  }
      
	  if (mTrendFlag) mTrend.trendData(chanv[id+1].mName,chanv[id+1].mTS->getStartTime(),
			 avgRhoSqPairwise[id][iOffset]);
	}
      }
      if (mLog && !emptyLog) logFile << endl;
    }

    if (mLog && !emptyLog) logFile.close();


    //****************************************************************//
    //  Produce eps output using native  ROOT graphics                //
    //****************************************************************//


    if (mEps) {
      gROOT->SetBatch();
      gErrorIgnoreLevel=2000;    // Do not produce annoying ROOT graphics warnings 

      // Determine lowest frequency band 
      if (minPlotFreq < 0) {
	if (mDebug) std::cerr << "# CorrMon: Minimum frequency set to 0" << std::endl;
	minPlotFreq=0;     
      }   
      nOffset=int(floor(minPlotFreq/(correlationWidth*deltaF)));


      // Maximum available frequency
      float maxAvailableFreq=nRho*correlationWidth*deltaF;
 
      if (maxPlotFreq >  maxAvailableFreq) {
	if (mDebug) std::cerr << "# CorrMon: Maximum available frequency " << maxAvailableFreq << std::endl;
	maxPlotFreq = maxAvailableFreq;     
      }
      nPlot=int(floor(maxPlotFreq/(correlationWidth*deltaF)))-nOffset;
	
      initialiseGraphics();
      addTitle();
      addChannelNames();
      typedef  TH1F* TH1F_ptr;
      corrGraph=new TH1F_ptr[nEnvChan]; 
      
      for (int iPad=0;iPad<2;iPad++) {
	initialisePad(corrPad[iPad]);
	for (int id=0;id<nEnvChan;id++) fillGraph(iPad, id);
	corrPad[iPad]->Modified();
	corrPad[iPad]->Draw();
	c->Update(); 
      }

      makeEPSFile();
    }
  } // End of mCount=nAveragingTimes loop

  dumpHTML(chanv[0].mName,htmlFile.c_str());


 
  mCount++;

  if (mCount > nAveragingTimes) {
    mPasses--;
    if (mDebug && mPasses>0) cerr << "# CorrMon: Pass finished - " << mPasses << " passes left" << endl;
    if (mPasses == 0) finish();
    mCount=1; // reset mCount
  }
  delete [] band;
}

CorrMon::~CorrMon() {
    //---------------------------------  Clean up at the end
  if (mSummary) summaryFile.close();

    for (int id=0 ; id<nChan ; id++) {
      mIn.rmChannel(chanv[id].mName);
      delete chanv[id].mTS;
      delete chanv[id].mFS;
    }
    delete [] chanv;

    for (int id=0;id<nEnvChan;id++) {
      delete [] rhoSqPairwise[id];
      delete [] avgRhoSqPairwise[id];
    }
    delete [] rhoSqPairwise;
    delete [] avgRhoSqPairwise;
    delete [] rhoSq;

    delete [] mA;
    delete [] mB;

    if (mDebug)  std::cerr << "# CorrMon: CorrMon finished" << std::endl;
}

void CorrMon::initialiseGraphics() {
  gROOT->Reset();
  gStyle->SetOptTitle(0); //Don't show title  
  gStyle->SetOptStat(0); // Don't show statistics
  gROOT->SetInterrupt(kTRUE);

  c = new TCanvas("CorrMon", "Interchannel Correlations",10,10,788,750); 
  c->SetHighLightColor(2);
  c->SetFillColor(33); // Blue-grey in the default ROOT palette
  c->SetBorderSize(2); 
  c->SetLeftMargin(0.12); 
  c->SetRightMargin(0.0);
  c->SetTopMargin(0.08); 
  c->SetBottomMargin(0.12); 
  c->SetEditable();

  corrPad[0]=new TPad("Pad0","Fullscale",0.025,0.6,0.975,0.99,21);
  corrPad[0]->Draw();
  corrPad[1]=new TPad("Pad1","Detail",0.025,0.2,0.975,0.57,21);
  corrPad[1]->Draw();
  corrPad[2]=new TPad("Pad2","Description",0.0,0.0,1.0,0.19,10);
  corrPad[2]->Draw();
} 

void CorrMon::addTitle() {
  corrPad[2]->cd();
  TPaveText *corrTitle = new TPaveText(0.025, 0.07, 0.975, 0.33,"NDC");
  corrTitle->SetFillColor(42);
  corrTitle->SetMargin(0.025);
  corrTitle->SetTextAlign(12);
  char title[128]="Interchannel Correlations with ";
  strcat(title,chanv[0].mName);
  corrTitle->AddText(title);       
  corrTitle->Draw();
}

void CorrMon::addChannelNames() {
  for (int id=0;id<nEnvChan;id++) {
    Float_t xText=0.05+0.24*(id-id%4)/4;  // 4 rows of names
    Float_t yText=0.85-0.15*(id%4);       // 4 columns of names
    TText *tt = new TText(xText,yText,chanv[id+1].mName);
    tt->SetTextColor(root_colors[id%N_ROOT_COLORS]);
    tt->SetTextSize(0.1);
    tt->Draw();
  }
}

void CorrMon::initialisePad(TPad *pcorrPad) {
  pcorrPad->cd(); 
  pcorrPad->SetFillColor(10); 
  pcorrPad->SetBorderMode(0);
  pcorrPad->SetGrid(); 
  pcorrPad->SetTicks();
  pcorrPad->SetFrameFillColor(10);
  pcorrPad->Draw();
}


void CorrMon::fillGraph(int iPad, int id) {
 
  corrGraph[id]=new TH1F(chanv[id+1].mName,chanv[id+1].mName,nPlot,0.0,maxPlotFreq);
  corrGraph[id]->SetMaximum(1.0);
  if (iPad==0)
    corrGraph[id]->SetMinimum(0.0);
  else	 
    corrGraph[id]->SetMinimum(lowLevelTrigger);
  corrGraph[id]->SetLineColor(id+2);     
  for (int i=0;i<nPlot;i++) {
    corrGraph[id]->Fill((i+0.5)*correlationWidth*deltaF,avgRhoSqPairwise[id][i]);
  }
  if (id==0) 
    corrGraph[id]->Draw("L"); // Draw points
  else
    corrGraph[id]->Draw("Lsame"); // Draw points on same plot
}


void CorrMon::makeEPSFile() {
  string epsFileName = startTimeString+".eps";
  c->Print(epsFileName.c_str());
}

void CorrMon::corrThreshold()
{ 
  //  do we threshold 
  for (int col=0;col<nEnvChan;col++) {
    if ( mB[col].MagSq() < threshold*mA[RC(col,col,nEnvChan)].Real()*mModX2Sum ) mB[col]=0.0; 
    for (int row=0;row<=col;row++) {  // as we only need give the upper triangular part of the matrix
      if ( mA[RC(row,col,nEnvChan)].MagSq() < threshold*mA[RC(row,row,nEnvChan)].Real()*mA[RC(col,col,nEnvChan)].Real() ) 
	mA[RC(row,col,nEnvChan)]=0.0;
    }        
  }
}

void CorrMon::writeRhoSq(float freq,float rho2)
{
  //  Write the list of rho squared values to file                         

  rhoSqOFS << freq << "\t\t" << rho2 << "\t\t";
  for (int chan=0;chan<nEnvChan;chan++) rhoSqOFS << rhoSqPairwise[chan] << " ";
  rhoSqOFS << std::endl;
}

void CorrMon::closeRhoSqFile()
{
    rhoSqOFS.close();
}

void CorrMon::rhoSqTrigger(int id,int iOffset)
{
	
  trig::TrigRslt t("Rho2", chanv[id+1].mName, 1);
  t.setTime(chanv[id+1].mTS->getStartTime());
  t.setPriority(trig::p_info);
  t.setDisposition(trig::d_metaDB);
  t.setDuration(chanv[id+1].mTS->getInterval());
  t.setIntensity(rhoSqPairwise[id][iOffset]);
  int rc = sendTrigger(t);
  if (rc && rc!=12 ) {
    if (mDebug) std::cerr << "Error " << rc << " logging Band " 
	      << chanv[id+1].mName << " trigger" << std::endl;
  }
}

int CorrMon::chanClean(int offset,FSeries *band)
{ 
  /*          NB f2c.h defines a data types complex and integer (= long int)     */
  /*                   which is used by clapack routines                         */

  /* Our equation is    y^p_i^* y^q_i r_q = x_i y^p_i*    */
  /* The lapack matrix convention is that the n x n matrix A_pq 
     is stored in a 1-dimensional array of size n^2 with
                A_pq = A[p+n*q]                           
     i.e we cycle fast down the rows of the first column 
     and then slowly across to the next column. Eg packing is as
                 ( 0   3   6 )
		 ( 1   4   7 )
		 ( 2   5   8 )                            */

  // set lapack parameters for chesv  

   CLAPACK_integer n = (CLAPACK_integer) nEnvChan; // Size of matrix 
   CLAPACK_integer lda = (CLAPACK_integer) nEnvChan;
   CLAPACK_integer ldb = (CLAPACK_integer) nEnvChan;
   CLAPACK_integer nrhs = (CLAPACK_integer) 1;    
   CLAPACK_integer *ipivot = new CLAPACK_integer[n]; 
   CLAPACK_integer lwork = (CLAPACK_integer)  nEnvChan;
   CLAPACK_complex *work = new CLAPACK_complex[lwork];

   char s = 'U';   // We give the upper triangular part of the matrix 

   CLAPACK_complex *R = new CLAPACK_complex[nEnvChan];
   CLAPACK_complex *A = new CLAPACK_complex[nEnvChan*nEnvChan];
  
   //         Call to the lapack routine chesv to solve the equation AR=B        
   //   Set up so that R holds the value B on input and the solution on output   
 
   for (int col=0;col<nEnvChan;col++) {
     R[col].r=mB[col].Real();
     R[col].i=mB[col].Imag();
     for (int row=0;row<=col;row++) {
       A[RC(row,col,nEnvChan)].r=mA[RC(row,col,nEnvChan)].Real();
       A[RC(row,col,nEnvChan)].i=mA[RC(row,col,nEnvChan)].Imag();
     }
   } 

  // Call to lapack routine chesv_  
	
   CLAPACK_integer info; 
		                                                  
   chesv_(&s,&n,&nrhs,A,&lda,ipivot,R,&ldb,work,&lwork,&info);

   if (info != 0) {
     std::cerr << "# " << __FILE__ << ": Non-zero completion code " << info << " returned by chesv" << std::endl;
   }

   // Calculate the significance measure |rho|^2  

   fComplex *mR = new fComplex[nEnvChan];

   for (int col=0;col<nEnvChan;col++) {
     mR[col]=fComplex(R[col].r,R[col].i);
   }

   fComplex rho2(0.0,0.0);
   for (int row=0;row<nEnvChan;row++) {
     rho2+=mR[row]*~mB[row];
   }    

  if ( mModX2Sum == 0 ) {
    std::cerr << "Signal channel identically 0, aborting ..." << std::endl;
    abort();
  }

  rho2/=mModX2Sum;
  if ( fabs(rho2.Imag()) > 0.00001 ) {
    std::cerr << "# CorrMon: Round-off problem: rho2.Imag() = " << rho2.Imag() << " (should be 0)" << std::endl; 
  }
   
  if ( rho2.Real() < threshold ) {  
     for (int row=0;row<nEnvChan;row++) mR[row]=0.0;
  }

  // Calculate the `cleaned' signal 
    
  fComplex *FSBand = new fComplex[correlationWidth];

  for (int k=0;k<correlationWidth;k++) {
    band[0].getData(correlationWidth,FSBand);
    cleanedFSData[k+offset]=FSBand[k];
  }

  for (int row=0;row<nEnvChan;row++) {  
    band[row+1].getData(correlationWidth,FSBand);
    for (int k=0;k<correlationWidth;k++) {
      cleanedFSData[k+offset]-=~mR[row]*FSBand[k];
    }
  }
  delete [] ipivot;
  delete [] work;
  delete [] R;
  delete [] A;
  delete [] mR;
  delete [] FSBand;
  return(info);
}


void CorrMon::readConfigFile(const char* pconfig) {

  //---------------------------------  Get parameters.
  ifstream ifs(pconfig, ios::in);
  if (ifs.bad()) {
    std::cerr << "CorrMon: Unable to open configuration file " << pconfig << std::endl;
    finish();
    return;
  }

  if (mDebug) std::cerr << "# CorrMon: Configuration file is " << pconfig << std::endl;

  while (1) {
    ifs.getline(descriptor, sizeof(descriptor));
    if (descriptor[0] == '#') continue;   // Ignore any preceeding comments
    if (mDebug) std::cerr << "# CorrMon: Descriptor is " <<  descriptor << std::endl;
    break;
  }

  ifs >> correlationWidth;
  if (mDebug) std::cerr << "# CorrMon: Correlation width = " << correlationWidth << std::endl;
     
  ifs >> dTime;   
  if (mDebug) std::cerr << "# CorrMon: Sample time = " << dTime << std::endl;

  //    window should be set to true if we want to window the data 
  ifs >> mWindow;
  if (mDebug) std::cerr <<  "# CorrMon: window = " << mWindow << std::endl;


  ifs >> nAveragingTimes;
  if (mDebug) std::cerr << "# CorrMon: Number of sample times averaged = " << nAveragingTimes << std::endl;

  ifs >> lowLevelTrigger;
  if (mDebug) std::cerr << "# CorrMon: Low level trigger threshold = " << lowLevelTrigger << std::endl;
  
  if (lowLevelTrigger < 0 || lowLevelTrigger >= 1) {
    if (mDebug) std::cerr << "# CorrMon:  Low level trigger threshold must be >=0 and < 1" << std::endl;
    abort();
  }

  ifs >> minPlotFreq;
  if (mDebug) std::cerr <<  "# CorrMon: Minimum Plotted Frequency Requested = " << minPlotFreq << std::endl;

  ifs >> maxPlotFreq;
  if (mDebug) std::cerr <<  "# CorrMon: Maximum Plotted Frequency Requested = " << maxPlotFreq << std::endl;

  //    clean should be set to 1 if we want to calculate the cleaned signal  
  ifs >> mClean;
  if (mDebug) std::cerr <<  "# CorrMon: clean = " << mClean << std::endl;

  ifs >> nChan;
  if (mDebug) std::cerr << "# CorrMon: Number of channels  (including signal) = " << nChan << std::endl;

 
  chanv=new Chan[nChan];

  for (int id=0 ; id<nChan ; id++) {
    ifs >> chanv[id].mName >> chanv[id].mRate;
    chanv[id].mPoint=int(ceil(dTime*chanv[id].mRate));
    chanv[id].mTS = new TSeries(Time(0), Interval(dTime), chanv[id].mPoint);
    chanv[id].mFS = new FSeries;
    chanv[id].mSpec = new FSpectrum;
    mIn.addChannel(chanv[id].mName, 0, &(chanv[id].mTS));
    if (mDebug) std::cerr << "# CorrMon: Added channel " << chanv[id].mName << "\t No of points :" << chanv[id].mPoint << std::endl;  
  }

 
  if (mClean) {
    cleanedSignal.mPoint=chanv[0].mPoint;
    cleanedSignal.mTS = new TSeries(Time(0), Interval(dTime), chanv[0].mPoint);
    cleanedSignal.mFS = new FSeries;   
    cleanedSignal.mSpec = new FSpectrum;
    cleanedFSData = new fComplex[chanv[0].mPoint];
  } 

  if (nTSBuffer > chanv[0].mPoint) nTSBuffer=chanv[0].mPoint;
  if (mDebug) cerr << "# CorrMon: dmtviewer TSeries buffer size " << nTSBuffer << endl;

  mPlotDecimation=chanv[0].mPoint/nTSBuffer;
  if (mDebug) cerr << "# CorrMon: dmtviewer TSeries decimation " << mPlotDecimation << endl;

  plotTS =  new TSeries(Time(0),Interval(dTime),nTSBuffer);
  plotSpec = new FSpectrum(*plotTS);
  plotCleanedTS =  new TSeries(Time(0),Interval(dTime),nTSBuffer);
  plotCleanedSpec = new FSpectrum(*plotTS);

  ifs.close();
}


void CorrMon::dumpHTML(const char* channel, const char* ofile)
{
//----------------------------------  Build an html document

    html::document doc("CorrMon statistics");
    // list<lineEntry>::iterator it;
    int row;
    char cbuf[256];


    doc.setBackgroundColor(html::color("wheat"));

    html::block titleBlk("center");
    titleBlk.lineBreak();
    LocalStr(Now(), cbuf, "%M %d, %Y %02H:%02N");
    html::text title(string("Correlation Monitor Statistics at: ") + string(cbuf));
    html::font fonttitle ("Helvetica");
    title.setFont (fonttitle);
    title.setColor (html::color (0, 0, 128));
    title.setSize (html::size (+4));
    titleBlk.addObject (title);
    titleBlk.lineBreak();
    html::text subtitle(string("Primary Channel:  ")+string(chanv[0].mName));
    subtitle.setSize (html::size (+3));
    titleBlk.addObject(subtitle);
    titleBlk.lineBreak();
    sprintf(cbuf,"%d",correlationWidth);
    subtitle.insertText(string("Correlation width:  ")+string(cbuf));
    titleBlk.addObject(subtitle);
    doc.addObject(titleBlk);
  

    html::table parTbl(string(" "));
    parTbl.addColumn(" ");
 
    parTbl.addColumn(" ");
    row = parTbl.addRow();
    row = parTbl.addRow();
    parTbl.insertData(row, 0, string("start GPS time:  "));
    parTbl.insertData(row, 1, long(startGPS));
    row = parTbl.addRow();
    parTbl.insertData(row, 0, string("current GPS time:&nbsp;&nbsp;"));
    parTbl.insertData(row, 1, long(currentGPS));
    row = parTbl.addRow();
    parTbl.insertData(row, 0, string("live time, sec:  "));
    parTbl.insertData(row, 1, long(liveGPS));
 
    doc.addObject(parTbl);

    doc.addObject (html::hline());
    doc.addObject (html::hline());

//----------------------------------  Add table for each Channel.

 
    for (int id=0; id<nEnvChan;id++) {

      html::table corrTbl("");
      corrTbl.addColumn("frequency");
      corrTbl.addColumn("&nbsp;&nbsp;&nbsp;mean correlation"); 
      corrTbl.addColumn("&nbsp;&nbsp;&nbsp;max correlation"); 
      corrTbl.addColumn("&nbsp;&nbsp;&nbsp;&nbsp;no of entries");
      corrTbl.addColumn("&nbsp;&nbsp;duty cycle");
 

      html::block subtitleBlk("center");
      subtitleBlk.lineBreak();
      subtitle.insertText("Correlations with "+string(chanv[id+1].mName));
      subtitleBlk.lineBreak();
      subtitleBlk.addObject(subtitle);
      doc.addObject(subtitleBlk);

      row = corrTbl.addRow();
      corrTbl.insertData(row, 0, "&nbsp;&nbsp;&nbsp;Hz");
      corrTbl.insertData(row, 1, "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp&nbsp;&nbsp&nbsp;&nbsp;[0,1]");
      corrTbl.insertData(row, 2, "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp&nbsp;&nbsp;[0,1]");
      sprintf(cbuf,"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Max %d",nAveragingTimes);
      corrTbl.insertData(row, 3, string(cbuf));
      corrTbl.insertData(row, 4, "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;%");

      double duty;
      for (int i=0 ; i < nRho ; i++) {

	if (monitorCount[id][i] > 0) {
	  row = corrTbl.addRow();
	  corrTbl.insertData(row, 0, frequencies[i].freqString);
	  sprintf(cbuf,"&nbsp;&nbsp;&nbsp;&nbsp;%13.2f",avgRhoSqPairwise[id][i]);	  
	  corrTbl.insertData(row, 1, cbuf);
	  sprintf(cbuf,"&nbsp;&nbsp;&nbsp;&nbsp;%13.2f",maxRhoSqPairwise[id][i]);	  
	  corrTbl.insertData(row, 2, cbuf);
	  sprintf(cbuf,"&nbsp;&nbsp;&nbsp;%8d",monitorCount[id][i]);
	  corrTbl.insertData(row, 3, cbuf);
	  duty = 100.*monitorCount[id][i]*dTime/liveGPS;
          corrTbl.insertData(row, 4, duty);
        }
      }

    doc.addObject(corrTbl);
    doc.addObject (html::hline());

   }




    html::table legTbl("Legend");

    legTbl.addColumn("");
    row = legTbl.addRow();
    legTbl.insertData(row, 0, 
     "Correlation - statistical measure of significance of line");
    row = legTbl.addRow();
    sprintf(cbuf,"duty cycle - fraction of time when correlation is above the threshold %9.2f",lowLevelTrigger);
    legTbl.insertData(row, 0,string(cbuf));
      //row = legTbl.addRow();
    // sprintf(cbuf,"                                         (THRESHOLD value %9.2f)",lowLevelTrigger);
      // legTbl.insertData(row, 0, cbuf);


    doc.addObject(legTbl);



//----------------------------------  Write the document

    ofstream f(string(ofile).c_str());
    html::writer w(f);
    doc.write(w);

}

void CorrMon::usage(void)
{
  std::cerr << "Command syntax is:" << std::endl;
  std::cerr << "\t CorrMon -conf <config>              see below"  << std::endl;
  std::cerr << "\t     Optional flags:"  << std::endl;
  std::cerr << "\t         [-verbose or -debug or -d]      produce verbose output to stderr" << std::endl;
  std::cerr << "\t         [-log]                          produce log file of values exceeding the given threshold" << std::endl;    
  std::cerr << "\t                                            (one per sampling time)" << std::endl;
  std::cerr << "\t         [-summary]                      produce a summary file of values exceeding the given threshold" << std::endl;
  std::cerr << "\t                                            one per (sampling time)x(no of sample times to average)" << std::endl;
  std::cerr << "\t         [-notrig or -notriggers]        suppress the generation of triggers" << std::endl;
  std::cerr << "\t         [-notrend]                      suppress the generation of trend files" << std::endl;
  std::cerr << "\t         [-eps]                          produce encapsulation postscript graphs of correlations" << std::endl;
  std::cerr << "\t                NB The -eps option is only foruse CorrMonRoot (the non-interactive version of CorrMon)" << std::endl;
  std::cerr << "\t         [-npass or -npasses] n          go through just n stretches of data of " << std::endl;
  std::cerr << "\t                                            length (sampling time)x(no of sample times to average)" << std::endl;
  std::cerr << "\t         [-onepass or -once]             go through just one stretch of data of " << std::endl;
  std::cerr << "\t                                            length (sampling time)x(no of sample times to average)" << std::endl;
  std::cerr << "\t                                            equivalent to npass=1 " << std::endl;
  std::cerr << "\t         [-help or -?]                     this message" << std::endl;
  std::cerr << "Configuration file syntax is:" << std::endl;
  std::cerr << "   # Comments (start with a #)" << std::endl;
  std::cerr << "   # More comments" << std::endl;
  std::cerr << "   ...  " << std::endl;
  std::cerr << "   # Last comment" << std::endl;
  std::cerr << "   Line 1:  <descriptor>" << std::endl;
  std::cerr << "   Line 2:  <correlation width>" << std::endl;
  std::cerr << "   Line 3:  <sampling time>" << std::endl;
  std::cerr << "   Line 4:  <window: 0=no, 1=yes>" << std::endl;
  std::cerr << "   Line 5:  <no of sample times to average>" << std::endl;
  std::cerr << "   Line 6:  <rho2 trigger threshold>" << std::endl;
  std::cerr << "   Line 7:  <lower frequency plotted (only applies to eps output)> " << std::endl;
  std::cerr << "   Line 8:  <upper frequency plotted (only applies to eps output)>" << std::endl;
  std::cerr << "   Line 9:  <should we clean: 0=no, 1=yes>" << std::endl;
  std::cerr << "   Line 10: <number of channels>" << std::endl;
  std::cerr << "   Line 11: <signal channel> <frequency> " << std::endl;
  std::cerr << "   Line 12: <first environmental channel> <frequency>" << std::endl;
  std::cerr << "   ..." << std::endl;
  std::cerr << "   Line 10+number of channels: <last environmental channel> <frequency>" << std::endl;
  std::cerr << std::endl;
  std::cerr << "   The fields above have the following meanings:" << std::endl;
  std::cerr << std::endl;
  std::cerr << "   <descriptor>                        A character string describing the run, used  for" << std::endl;
  std::cerr << "                                           naming any output files" << std::endl;
  std::cerr << "   <number of channels>                Number of channels including the signal " << std::endl;
  std::cerr << "   <sample time>                       The period in seconds over which correlations are sought." << std::endl;
  std::cerr << "   <signal/environmental channel>      The name of the signal/environmental " << std::endl;
  std::cerr << "                                           channel between which correlations are sought." << std::endl;
  std::cerr << "   <frequency>                         The sample rate of the channel" << std::endl;
  std::cerr << "   <clean>                             Set to 1 if you want a best estimate of the `true' spectrum" << std::endl;
  std::cerr << "                                           based on a knowledge of correlations gleaned, 0 otherwise." << std::endl;

  finish();
}
