/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef PDnMon_HH
#define PDnMon_HH

#include "AlarmClient.hh"
#include "DatEnv.hh"
#include "MonServer.hh"
#include "FixedLenTS.hh"
#include "Histogram1.hh"
#include "OperStateCondList.hh"
#include "Trend.hh"
#include "TrigPar.hh"
#include "autopipe.hh"
#include <string>
#include <list>

//
//    Data Monitor template class.

/**  @name PDnMon
  *  @memo Template of a data environment (DatEnv) based monitor.
  *  The PDnMon monitor is a template for a data monitor program.
  *  PDnMon makes a simple calculation of the standard deviation of
  *  a specified channel (by default H0:PEM-LVEA_SEISX), serves the result 
  *  for display by the DMT viewer and stores the result as a minute trend.
  *
  *  {\bf {\Large Running PDnMon}}
  *
  *  The PDnMon command line syntax is:
  *  \begin{verbatim}
        PDnMon -conf <file>
     \end{verbatim}
  *  The #<number># argument specifies the number of data strides to process
  *  before terminating the program. The #<time># argument specifies the data
  *  stride. If "-dvIndex" is specified, a trend index file will be written
  *  to '\$DMTOUTPUT/channel.cnf'.
  *
  *  {\bf {\Large Configuration File}}
  *
  *  PDnMon uses a configuration file.
  *
  *  {\bf {\Large PDnMon Output}}
  *
  *  {\bf Trend channels}
  *  PDnMon produces a minute trend of the sigma value called 
  *  #<channel>_sigma#
  *
  *  {\bf Monitor data channels}
  *  PDnMon publishes a time series history of the sigma values 
  *  with the name #<channel>_sigma#
  *
  *  @author John Zweizig; Caltech/LIGO
  *  @version 1.1; Last modified August 12, 2004
  */
//{
//}

class PDnMon : public DatEnv, MonServer {
public:
    PDnMon(int argc, const char *argv[]);  // Constructor
    ~PDnMon(void);                         // Destructor
    void ProcessData(void);                     // Process 1 data stride
    void Attention(void);                       // Handle Message signal.
    bool configure(const std::string& file);

    /**  Print status
      */
    void printStat(void) const;

private:

    /** Class to hold a single redundant sub-channel.
     */
    class subchan {
    public:
	/**  Constructor - all that's needed is a channel name.
	  */
	subchan(const std::string& name, Trend& trend);

	/**  Destructor.
	  */
	~subchan(void);

	/**  Get the constant data offset.
	  */
	double getBias(void) const;

	/**  Get the channel data.
	  */
	const TSeries& getData(void) const;

	/**  Get the channel name.
	  */
	const char* getName(void) const;

	/**  Get the current scale factor.
	  */
	double getScale(void) const;

	/**  Get the scaled data.
	  */
	TSeries getScaledData(void) const;

	/**  Print status
	 */
	void printStat(std::ostream& out) const;

	/**  Read the channel data from the dacc into an internal Time series.
	  */
	void readChan(const Dacc& dacc);

	/**  Calculate scaling factor and offset from the nominal signal.
	 */
	const std::string& refQuadName(void) const;

	/**  Reset the filtering.
	  */
	void reset(void);

	/**  Calculate scaling factor and offset from the nominal signal.
	 */
	void scale(const TSeries& ts);

	/**  Set debugging print level.
	 */
	void setDebug(int lvl) {_debug = lvl;}

	/**  Calculate scaling factor and offset from the nominal signal.
	 */
	void setFilter(const Pipe& pipe);

	/**  Calculate scaling factor and offset from the nominal signal.
	 */
	void setQuadName(const std::string& name);

	/**  Calculate scaling factor and offset from the nominal signal.
	 */
	void setViewer(MonServer& ms, Interval histl);

	/**  Calculate scaled offsets squared.
	 */
	TSeries sqoff(const TSeries& ts);

    private:
	std::string _name;
	std::string _quadname;
	int         _debug;
	std::string _troot;
	Trend&      _trend;
	TSeries     _raw;
	TSeries     _quad;
	mutable TSeries _scratch;
	double      _off;
	double      _scale;
	double      _phase;
	auto_pipe   _pipe;
	auto_pipe   _qpipe;
	FixedLenTS  _bhist;       //  bias history.
	FixedLenTS  _shist;       //  scale history.
	TSeries     _terms;
    };

    /** Class to hold a redundant channel group.
     */
    class group {
    public:
	/**  Constructor - all that's needed is a channel name.
	  */
	group(const std::string& name, Dacc& dacc, Trend& trend);

	/**  Destructor.
	  */
	~group(void);

	/**  Add a subchannel.
	  */
	void addSubChan(const std::string& name);
	void addSubChan(const subchan& sb);

	/**  Generate alarms as indicated
	  */
	void alarm(AlarmClient& ac);

	/**  Calculate the average value, calculate a chi-square for
	  *  each sample, and look for samples over threshold.
	  */
	void calculate(void);

	/**  Get the OSC condition name.
	  */
	const char* getCondition(void) const;

	/**  Get the group name.
	  */
	const char* getName(void) const;

	/**  Print status
	 */
	void printStat(std::ostream& out) const;

	/**  Set debugging print level.
	 */
	void setCondition(const std::string& cond) {_osc_cond = cond;}

	/**  Set debugging print level.
	 */
	void setDebug(int lvl) {_debug = lvl;}

	/**  Set the filter.
	 */
	void setFilter(const Pipe& pipe);

	/**  Set the threshold.
	 */
	void setThreshold(double thresh);

	/**  Calculate scaling factor and offset from the nominal signal.
	 */
	void setViewer(MonServer& ms, Interval histl);


    private:
	typedef std::list<subchan>      sublist;
	typedef sublist::iterator       subchan_iter;
	typedef sublist::const_iterator const_subchan_iter;

    private:
	std::string _name;
	std::string _osc_cond;
	Dacc&       _dacc;
	int         _debug;
	std::string _troot;
	Trend&      _trend;
       	sublist     _list;
	double      _weight;
	long        _avgcnt;
	TSeries     _avg;
	double      _thresh;
	double      _scale_alarm;
	double      _bias_alarm;
	Time        _current;
	Time        _segstart;
	auto_pipe   _pipe;
	FixedLenTS  _chist;       //  rate history.
	FixedLenTS  _rhist;       //  rate history.
	Histogram1  _histo;
    };

private:
    typedef std::list<group>          grouplist;
    typedef grouplist::iterator       group_iter;
    typedef grouplist::const_iterator const_group_iter;

private:
    int       maxFrame;       //  Number of frames to process
    Interval  mStep;          //  Time stride.
    TrigPar   mDict;
    OperStateCondList mOSC;
    Trend     mTrend;         //  Trend writer
    grouplist mList;
    AlarmClient mAlarm;
};


//======================================  PDnMon::subchan inline methods
inline const char* 
PDnMon::group::getName(void) const {
    return _name.c_str();
}

inline const char* 
PDnMon::group::getCondition(void) const {
    return _osc_cond.c_str();
}

inline double 
PDnMon::subchan::getBias(void) const {
    return _off;
}

inline const TSeries& 
PDnMon::subchan::getData(void) const {
    return _raw;
}

inline const char* 
PDnMon::subchan::getName(void) const {
    return _name.c_str();
}

inline double 
PDnMon::subchan::getScale(void) const {
    return _scale;
}

inline const std::string& 
PDnMon::subchan::refQuadName(void) const {
    return _quadname;
}

#endif     //  PDnMon_HH
