#ifndef TRIGMGRCLIENT_HH
#define TRIGMGRCLIENT_HH

#include "TrigWriter.hh"
#include "lmsg/MsgTypes.hh"
#include <string>

namespace trig {
  class TrigBase;
  class Segment;
}

/**  The TrigClient class provides the client API for the Trigger Manager
  *  server. The trigger manager maintains a registry of monitors that 
  *  may produce triggers. The client API requests that it be added to the
  *  registry with the enroll method. The client may then sends one or more 
  *  triggers to the trigger manager with the #sendTrigger()# method. The
  *  trigger data may be routed to an xml file instead by setting the 
  *  TRIGGER_FILE environmwnt variable.
  *  @memo Trigger Client API
  *  @author John G. Zweizig
  *  @version 1.4; Modified November 24, 2003
  */
class TrigClient {
public:
  /**  Construct a trigger client.
    *  @memo Construct a trigger client.
    */
  TrigClient(void);

  /**  Construct a trigger client with the specified mode.
    *  @memo Construct a trigger client.
    */
  TrigClient(trig::TrigWriter::trig_mode mode);

  /**  Destroy a trigger client.
    *  @memo Destroy a trigger client.
    */
  virtual ~TrigClient(void);

  /**  The process is registered with the trigger manager. On successful
    *  enrolment, the trigger manager returns a unique process identifier
    *  which is used to identify the process in subsequent requests.
    *  @memo Enroll process with the Trigger Manager.
    *  @return Error code enumerated in lmsg::Errorlist.hh.
    */
  lmsg::error_type enroll(trig::TrigWriter::trig_mode mode=trig::TrigWriter::kNone);

  /**  Get the debug printout level.
    *  @memo get the debug level.
    */
  lmsg::index_type getDebug(void) const;

  /**  Send a segment to be logged by the trigger manager.
    *  @memo Log a Segment with the trigger manager.
    *  @return Error code enumerated in lmsg::Errorlist.hh.
    *  @param segm Segment to be logged in the metadatabase.
    */
  lmsg::error_type sendSegment(const trig::Segment& segm);

  /**  Send a trigger to the trigger manager.
    *  @memo Log a trigger with the trigger manager.
    *  @return Error code enumerated in lmsg::Errorlist.hh.
    *  @param trigger Trigger to be logged in the metadatabase.
    */
  lmsg::error_type sendTrigger(const trig::TrigBase& trigger);

  /**  Set the debug level.
    *  @memo Set the debug level.
    *  @param level New debugging level.
    */
  void setDebug(lmsg::index_type level);

  /**  Set the table output file name and the mode to kWriter. If the 
    *  mode has been previously set to kMgr (logging via the trigger 
    *  manager) this request is ignored.
    *  @memo Set the table file.
       @param name Table file name
    */
  void setTableFile(const char* name);

private:
  /**  This function writes out anydata that is stored in the internal 
    *  database tables.
    */
  void writeTables(void);

private:
  lmsg::index_type  mDebug;
  trig::TrigWriter* mWriter;
  std::string       mTableFile;
};

inline lmsg::index_type 
TrigClient::getDebug(void) const {
    return mDebug;
}


#endif  //   TRIGMGRCLIENT_HH
