#include "TrigTable.hh"
#include "TrigBase.hh"
#include "TrigProc.hh"
#include "Segment.hh"
#include <fstream>
#include <stdlib.h>

using namespace xsil;
using namespace std;

namespace trig {

//======================================  Pack N decimal digits into N/2 bytes
static void
packKey(int len, const char* in, unsigned char* out) {
    for (int j=0 ; j < 2*len ; j+=2) {
        *out++ = ((in[j] & 0x0f)<<4) + (in[j+1] & 0x0f);
    }
}

//======================================  Define the trigger table.
TrigTable::TrigTable(bool alphaPID) 
  : MetaTable("gds_trigger"), binarydata(&binarydata_length, data),
    proc_id_length(nProcessKeyBytes), process_id(&proc_id_length, processID),
    altPID(alphaPID)
{
    if (altPID) defineColumn("process_id", &PIDstr, true);
    else        defineColumn("process_id", &process_id);
    defineColumn("name",              &name);
    defineColumn("subtype",           &subtype);
    defineColumn("ifo",               &ifo);
    defineColumn("start_time",        &start_time);
    defineColumn("start_time_ns",     &start_time_ns);
    defineColumn("duration",          &duration);
    defineColumn("priority",          &priority);
    defineColumn("disposition",       &disposition);
    defineColumn("size",              &size);
    defineColumn("significance",      &significance);
    defineColumn("frequency",         &frequency);
    defineColumn("bandwidth",         &bandwid);
    defineColumn("time_peak",         &peak_time);
    defineColumn("time_average",      &avg_time);
    defineColumn("time_sigma",        &time_sigma);
    defineColumn("freq_peak",         &freq_peak);
    defineColumn("freq_average",      &freq_avg);
    defineColumn("freq_sigma",        &freq_sigma);
    defineColumn("noise_power",       &noise_power);
    defineColumn("signal_power",      &signal_power);
    defineColumn("pixel_count",       &pixel_count);
    defineColumn("confidence",        &confidence);
    defineColumn("binarydata",        &binarydata);
    defineColumn("binarydata_length", &binarydata_length);
    defineColumn("event_id",          this);
}

//======================================  Delete the trigger table
TrigTable::~TrigTable(void) {
    MetaTable::clear();
}

//======================================  Fill the trigger table.
void
TrigTable::addRow(const trig::TrigBase& t) {
    start_time    = t.getTime().getS();
    start_time_ns = t.getTime().getN();
    priority      = 0;
    disposition   = t.getDisposition();
    if (altPID) PIDstr = t.getProcess();
    else        packKey(nProcessKeyBytes, t.getProcess(), processID);
    name          = t.getID();
    subtype       = t.getSubID();
    ifo           = t.getIFOs();
    duration      = t.getDt();
    size          = t.getIntensity();
    significance  = t.getSignificance();
    frequency     = t.getFrequency();
    bandwid       = t.getBandwidth();
    peak_time     = t.getPeakOffset();
    avg_time      = t.getAvgOffset();
    time_sigma    = t.getTimeSigma();
    freq_peak     = t.getFreqPeak();
    freq_avg      = t.getFreqAvg();
    freq_sigma    = t.getFreqSigma();
    noise_power   = t.getNoisePower();
    signal_power  = t.getSignalPower();
    pixel_count   = t.getPixelCount();
    confidence    = t.getConfidence();
    binarydata_length = t.getDataLength();
    t.getData(0, binarydata_length, (char*)data);
    putRow();
}

//======================================  Check that the pointers are OK
void
TrigTable::check(const char* title) const {
    MetaTable::check(title, sizeof(TrigTable));
}

//======================================  Define the process table.
ProcTable::ProcTable(bool alphaPID) 
  : MetaTable("process"), proc_id_length(nProcessKeyBytes), 
    process_id(&proc_id_length, processID), altPID(alphaPID)
{
    setColumns();
}

//======================================  Define the columns in the MetaTable
void
ProcTable::setColumns(void) {
    MetaTable::clear();
    defineColumn("program",        &program);
    defineColumn("version",        &version);
    defineColumn("cvs_repository", &cvsname);    
    defineColumn("cvs_entry_time", &cvstime);    
    defineColumn("comment",        &comment);
    defineColumn("node",           &node);
    defineColumn("username",       &username);
    defineColumn("unix_procid",    &unix_procid);
    defineColumn("start_time",     &start_time);
    defineColumn("end_time",       &end_time);
    if (altPID) defineColumn("process_id", &PIDstr, true);
    else        defineColumn("process_id", &process_id);
    defineColumn("ifos",           &ifo);
}

//======================================  Destroy the process table
ProcTable::~ProcTable(void) {
    MetaTable::clear();
}

//======================================  Fill the process table.
void
ProcTable::addRow(const trig::TrigProc& p) {
    program     = p.getName();
    version     = p.getVersion();
    cvsname     = p.getSource();
    cvstime     = p.getModTime();
    comment     = p.getTitle();
    node        = p.getNode();
    username    = p.getAuthor();
    unix_procid = p.getProcID();
    start_time  = p.getStartTime();
    end_time    = p.getEndTime();
    if (altPID) PIDstr = p.getProcessID();
    else        packKey(nProcessKeyBytes, p.getProcessID(), processID);
    ifo         = p.getIFOs();
    putRow();
}

//======================================  Check that the pointers are OK
void
ProcTable::check(const char* title) const {
    MetaTable::check(title, sizeof(ProcTable));
}

//======================================  Define the trigger table.
SegTable::SegTable(bool alphaPID)
  : MetaTable("segment"), proc_id_length(nProcessKeyBytes),
    process_id(&proc_id_length, processID), altPID(alphaPID)
{
    defineColumn("segment_group", &group);
    defineColumn("version",       &version);
    defineColumn("start_time",    &start_time);
    defineColumn("start_time_ns", &start_time_ns);
    defineColumn("end_time",      &end_time);
    defineColumn("end_time_ns",   &end_time_ns);
    if (altPID) defineColumn("process_id", &PIDstr, true);
    else        defineColumn("process_id", &process_id);
}

//======================================  Delete the trigger table
SegTable::~SegTable(void) {
    MetaTable::clear();
}

//======================================  Fill the trigger table.
void
SegTable::addRow(const trig::Segment& s) {
    group         = s.getGroup();
    version       = s.getVersion();
    start_time    = s.getStartTime().getS();
    start_time_ns = s.getStartTime().getN();
    end_time      = s.getEndTime().getS();
    end_time_ns   = s.getEndTime().getN();
    if (altPID) PIDstr = s.getProcess();
    else        packKey(nProcessKeyBytes, s.getProcess(), processID);
    putRow();
}

//======================================  Check that the pointers are OK
void
SegTable::check(const char* title) const {
    MetaTable::check(title, sizeof(SegTable));
}

//=====================================================================
//
//   The mySql database needs a few different tables
//
//=====================================================================

DQSegDef::DQSegDef(void) 
  : mVersion(0)
{
}

DQSegDef::DQSegDef(const Segment& seg) {
    mProcID  = seg.getProcess();
    mIfos    = seg.getIfos();
    mName    = seg.getGroup();
    mVersion = seg.getVersion();
    mComment = seg.getComment();
}

bool 
DQSegDef::operator==(const DQSegDef& x) const {
    return mIfos == x.mIfos && mName == x.mName && mVersion == x.mVersion;
}

int 
DQSegDef::getVersion(void) const {
    return mVersion;
}

const string& 
DQSegDef::refProcess(void) const {
    return mProcID;
}

const string& 
DQSegDef::refDefinerID(void) const {
    return mDefID;
}

const string& 
DQSegDef::refIfos(void) const {
    return mIfos;
}

const string& 
DQSegDef::refName(void) const {
    return mName;
}

const string& 
DQSegDef::refComment(void) const {
    return mComment;
}

void 
DQSegDef::setDefinerID(const std::string& defid) {
    mDefID = defid;
}

//======================================  Define the MySql segment table
DQSegTable::DQSegTable(void) 
  : MetaTable("segment")
{
    defineColumn("segment_id", &segid, true);
    defineColumn("start_time", &start_time);
    defineColumn("end_time",   &end_time);
    defineColumn("active",     &active);
    defineColumn("process_id", &procid, true);
}

//======================================  Delete the trigger table
DQSegTable::~DQSegTable(void) {
    MetaTable::clear();
}

//======================================  Fill the trigger table.
void
DQSegTable::addRow(const trig::Segment& s) {
    start_time = s.getStartTime().getS();
    end_time   = s.getEndTime().getS();
    active     = s.getActivity();
    segid      = s.getSegID();
    procid     = s.getProcess();
    putRow();
}

//======================================  Check that the pointers are OK
void
DQSegTable::check(const char* title) const {
    MetaTable::check(title, sizeof(DQSegTable));
}

//======================================  MySql segment definer table
DQSegDefTable::DQSegDefTable(void) 
  : MetaTable("segment_definer")
{
    defineColumn("process_id",     &procid, true);
    defineColumn("segment_def_id", &defid, true);
    defineColumn("run",            &run);
    defineColumn("ifos",           &ifos);
    defineColumn("name",           &name);
    defineColumn("version",        &version);
    defineColumn("comment",        &comment);

    //----------------------------------  Fill in the run name
    run = "Sn";
    const char* home = getenv("HOME");
    if (!home)  home = ".";
    string file(home);
    file += "/pars/RunID";
    ifstream ridfile(file.c_str(), ios::in);
    if (!ridfile) return;
    ridfile >> run;
}

//======================================  Delete the trigger table
DQSegDefTable::~DQSegDefTable(void) {
    MetaTable::clear();
}

//======================================  Fill the trigger table.
void
DQSegDefTable::addRow(const DQSegDef& d) {
    procid  = d.refProcess();
    defid   = d.refDefinerID();
    ifos    = d.refIfos();
    name    = d.refName();
    version = d.getVersion();
    comment = d.refComment();
    putRow();
}

//======================================  Check that the pointers are OK
void
DQSegDefTable::check(const char* title) const {
    MetaTable::check(title, sizeof(DQSegDefTable));
}

//======================================  MySql segment definer table
DQSegMapTable::DQSegMapTable(void) 
  : MetaTable("segment_def_map")
{
    defineColumn("segment_id",     &segid,  true);
    defineColumn("segment_def_id", &defid,  true);
    defineColumn("process_id",     &procid, true);
}

//======================================  Delete the trigger table
DQSegMapTable::~DQSegMapTable(void) {
    MetaTable::clear();
}

//======================================  Fill the segment map
void
DQSegMapTable::addRow(const trig::Segment& s, const trig::DQSegDef& d) {
    addRow(s.getSegID(), d.refDefinerID(), s.getProcess());
}

void 
DQSegMapTable::addRow(const string& s, const string& d, const string& p) {
    segid  = s;
    defid  = d;
    procid = p;
    putRow();
}


//======================================  Check that the pointers are OK
void
DQSegMapTable::check(const char* title) const {
    MetaTable::check(title, sizeof(DQSegMapTable));
}

} // namespace trig
