/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef TRIG_TRIGWRITER_HH
#define TRIG_TRIGWRITER_HH

#include "lmsg/MsgTypes.hh"
#include "TrigProc.hh"
#include <string>


namespace trig {

  class Segment;
  class TrigBase;

  /**  The TriggerWriter class is an API for writing triggers to various 
    *  trigger output databases and formats. It is used by the TrigClient 
    *  class. The usage mode is as follows:
    *  \begin{ennumerate}
    *  \item Construct the TrigWriter nased class with e.g. 
    *        mWriter = new LdasDBWriter;
    *  \item Set the default processID with:
    *       mWriter->setProcess(const trig::TrigProc&);
    *  \item Write segments/triggers with 
    *       mWriter->addSegment(const trig::Segment&)
    *       or with
    *       addTrigger(const trig::TrigBase&)
    *  \item The data are written to a file explicitly with:
    *       write(const std::string& file);
    *  \end{ennumerate}
    *  @memo Trigger writer API
    *  @author J. Zweizig
    *  @version 1.0; Last Modified September 23, 2005
    */
  class TrigWriter {
  public:
    enum trig_mode {
      kNone, 
      kMgr,
      kWriter,
      kSegWrt
    };

  public:
      /**  All local data are initialized to their default values.
        *  @memo Default constructor.
        */
      TrigWriter(void);

      /**  Close out write anddelete all local data.
        *  Default constructor.
	*/
      virtual ~TrigWriter(void);

      /**  The trigger is written to the trigger output stream and associated 
        *  with the specified process ID. In buffered writers, the trigger 
	*  (and process ID if unique) are entered in the writer tables until
	*  the buffer is flushed by a write command.
	*  @memo Write a trigger.
	*/ 
      virtual lmsg::error_type addTrigger(const TrigBase& t, const TrigProc& p)=0;

      /**  Write a trigger associated with the default processID (as 
        *  specified by the setProcess method).
	*  @memo Write a trigger.
	*/
      lmsg::error_type addTrigger(const TrigBase& t);

      /**  The segment is written to the segment output stream and associated 
        *  with the specified process ID. In buffered writers, the segment
	*  (and process ID if unique) are entered in the writer tables until
	*  the buffer is flushed by a write command.
	*  @memo Write a segment.
	*/ 
      virtual lmsg::error_type addSegment(const Segment& s, const TrigProc& p)=0;

      /**  Write a segment associated with the default processID (as 
        *  specified by the setProcess method).
	*  @memo Write a segment.
	*/
      lmsg::error_type addSegment(const Segment& s);

      /**  Clear the internal trigger and process tables.
        */
      virtual void clear(void) = 0;

      /**  Test whether the trigger and segment lists are empty.
        */
      virtual bool empty(void) const;

      /**  Get the debug message level.
        */
      virtual lmsg::index_type getDebug(void) const;

      /**  Get the trigger writer mode
        */
      virtual trig_mode getMode(void) const = 0;

      /**  Get the number of segments currently buffer by the writer.
        */
      virtual int getNSegs(void) const;

      /**  Get the number of trigggers currently buffered by the writer.
        */
      virtual int getNTrigs(void) const;

      /**  Return a constant reference to the current default process
        *  descriptor.
	*/
      const TrigProc& refProcess(void) const;

      /**  Set the debug message level.
       */
      virtual void setDebug(lmsg::index_type d);

      /**  Set the default process descriptor.
        */
      virtual lmsg::error_type setProcess(const TrigProc& p);

      /**  Flush the writer buffers of any stored segments, triggers or
        *  process entries.
	*/
      virtual lmsg::error_type write(const std::string& file) const = 0;

  protected:
      /**  Return a non-constane reference to the default Process descriptor.
        */
      TrigProc& refProcess(void);

  private:
      lmsg::index_type mDebug;
      TrigProc mProcess;
  };

} // namespace trig

inline lmsg::error_type
trig::TrigWriter::addTrigger(const TrigBase& t) {
    return addTrigger(t, mProcess);
}

inline lmsg::error_type
trig::TrigWriter::addSegment(const Segment& s) {
    return addSegment(s, mProcess);
}

inline trig::TrigProc&
trig::TrigWriter::refProcess(void) {
    return mProcess;
}

inline const trig::TrigProc&
trig::TrigWriter::refProcess(void) const {
    return mProcess;
}

#endif // TRIG_TRIGWRITER_HH
