#ifndef DACCIN_HH
#define DACCIN_HH

#include "Time.hh"
#include "Interval.hh"
#include "FileList.hh"
#include "FSeries.hh"
#include <string>
#include <iosfwd>

#define DACC_ONLDEV "/online/"

#ifndef __CINT__
  #include "framecpp/Types.hh"
  #include "framecpp/FrRawData.hh"
  #include "framecpp/FrStatData.hh"
  #include "framecpp/FrameH.hh"
  typedef FrameCPP::FrRawData::const_firstAdc_iterator AdcDataIterator;
  typedef FrameCPP::FrameH::const_procData_iterator    ProcDataIterator;
  typedef FrameCPP::FrameH::const_simData_iterator     SimDataIterator;
#endif

class DVector;
class FrStatDataRef;

/**  DaccIn methods provide an input interface to frame files.
  *  @memo DMT frame input class.
  *  @author  John G. Zweizig
  *  @version 1.2; Last modified May 26, 2004
  */
class DaccIn {
  public:
    //--------------------------------------  Constructors and destructors.
    /**  Default (Null) constructor.
      */
    DaccIn(void);

#ifndef __CINT__
    /**  Construct using an already allocated frame reader.
      */
    explicit DaccIn(FrameCPP::IFrameStream* reader);
#endif

    /**  Destructor.
      */
    virtual ~DaccIn();

    //---------------------------------------  Status manipulation
    /**  This parameter controls the fraction of the data read into 
      *  the shared memory to be copied to the data stream and what
      *  part of the shared memory resources can be held allocated the 
      *  stream. 'nbuf' == 0 indicates data is taken as requested.
      *  'nbuf' == -1 indicates that all data written to the shared 
      *  memory must be copied to the data stream. 'nbuf'==N indicates
      *  that up to N buffers (frames) of data should be held until the 
      *  stream is ready for them.
      *  @memo Set the number of online buffers to reserve.
      *  @param N Number of buffers to reserve, or -1 for all.
      */
    void setBuffer(int N);

    /**  If the debug flag is set, explanatory messages will be printed 
      *  when an error is encountered.
      *  @memo Set the debug flag.
      */
    void setDebug(int debug=0);

    /**  Set the read mode.
      *  @memo Set the readmode.
      */
    void setTOCMode(bool toc=true);

    //--------------------------------------  File list manipulation
    /**  One or more file names are added to the file list. Wilcard
      *  characters ('*'s) can be used to add several files simultaneously.
      *  The file names are sorted to keep them in an increaseing sequence.
      *  @memo Add a file to the file list.
      *  @param File Name of the file to be added to the input list.
      */
    void addFile(const char* File);

    /**  One or more frame file descriptors are read from a text file and 
      *  added to the list of frames to be processed. Each line of the 
      *  text file must contain a single file specifier like the those 
      *  accepted by addFile().
      *  @memo Add a file to the file list.
      *  @param File Name of the file to be added to the input list.
      */
    void addFileList(const char* File);

    /**  The current file is closed and its name is removed from the 
      *  file list. Any open shared memory partition is released.
      *  @memo Close the current file.
      */
    void close(void);

    /**  The first file in the file list is opened. If a file is already 
      *  open, it will be closed, and the next file will be opened. If the 
      *  file name is of the form "/online/<partname>" the input stream is 
      *  connected to the the shared memory partition specified by <partname>.
      *  @memo Open a file.
      *  @return Zero if open was sucessful.
      */
    int open(void);

    //----------------------------------  Data processing.
    /**  The current data frame is released. In Online data mode, the 
      *  shared memory partition buffer containing the frame is also 
      *  released. No new frame is read.
      *  @memo Release the current data frame.
      */
    void endFrame(void);

    /**  The current data frame is deleted and a new frame is read. 
      *  If the input file is not open or if an error occurs on reading
      *  the current file, the file is close and a new file is opened.
      *  @memo Read in a new data frame.
      *  @return Zero if valid frame read in, otherwise -1.
      */
    int nextFrame(void);

    //---------------------------------  Accessors
    /**  Returns the current frame time interval;
      *  @memo Get the Current time.
      */
    Interval getDt(void) const;

    /**  Returns the level of debug information that was requested.
      *  @memo Get the debug flag.
      *  @return Debug print level.
      */
    int getDebug(void) const {return mDebug;}

    /**  Returns a pointer to a string containing the current input file 
      *  name. If no file is open, a pointer to the next file name is 
      *  returned.
      *  @memo Get the input file name.
      */
    const char* getFile(void) const;

    /**  Get the time and comment of the specified history record.
      *  @memo Get history data.
      *  @param name History record name.
      *  @param t    History record time (returned);
      *  @param com  History record comment.
      *  @return true if history found.
      */
    bool getHistory(const std::string& name, Time& t, std::string& com) const;

#ifndef __CINT__
    /**  Returns a pointer to the current frame header.
      *  @memo Get a pointer to the current frame header.
      */
    const FrameCPP::FrameH* getFrame(void) const {return mFrame;}

    /**  An identifier of the form #<ifo-id>-<GPS-time>.F# is generated for 
      *  the current Frame and copied to the specified string.
      *  @memo Copy a Frame ID into the specified string.
      *  @return string containing the current frame identifier.
      */
    std::string getFrameID(void);
#endif

    /**  Fetch all specified static data.
      *  @memo Get a static data.
      *  @param name FrStat name.
      *  @param det  Detector name.
      *  @param startGps  Start time of static validity.
      *  @param endGps End time of static validity.
      *  @return Number of static structures read.
      */
    int getStaticData(const std::string& name=".*", 
		      const std::string& det=".*", 
		      const Time& startGps=Time(0), 
		      const Time& endGps=Time(0));
  
    /**  Get a frequency series (FSeries) from a previously read static 
      *  structure. Note that getStaticData must have been used previously 
      *  to insure that the necessary static data reside in memory.
      *  @memo Get a static fseries.
      *  @param name FrStat name.
      *  @param gps  Time for of static validity.
      *  @return FSeries.
      */
    FSeries getStaticFSeries(const std::string& name, const Time& gps);
  
    /**  Get a time series (TSeries) from a previously read static structure.
      *  Note that getStaticData must have been used previously to insure that
      *  the necessary static data reside in memory.
      *  @memo Get a static TSeries.
      *  @param name FrStat name.
      *  @param gps  Time for of static validity.
      *  @return TSeries.
      */
    TSeries getStaticTSeries(const std::string& name, const Time& gps);

    /**  Returns the start time of the current (or previous) frame. If 
      *  the Dacc has no currently valid frame, getTime returns the start 
      *  time of the last frame read. If no frames have been read, getTime 
      *  returns Time(0). 
      *  @memo Get the Start time.
      *  @return start time of current or previous frame.
      */
    Time getTime(void) const {return mT0;}

    long getTotalFrames(void) const {return mTotalFrames;}

    /**  Returns the time at which data collection is to start.
      *  @memo Get the Start time.
      */
    bool haveFrame(void) const {return (mFrame != 0);}

    /**  Test whether the data are from an online shared memory partition.
      *  @memo Test if online data.
      *  @return True if DaccIn input source is a shared memory partition.
      */
    bool isOnline(void) const {return (mSource==s_online);}
    bool isOpen(void) const {return mReader;}

    /**  Reference a file list
      */
    std::ostream& listFiles(std::ostream& out) const;
    FileList& refList(void);
    const FileList& refList(void) const;

    /**  Wait for data or a signal. waitData returns when a data buffer is
      *  found in the online partition or when a signal is caught. If the
      *  data are available, the return value is true. If #test_only# is 
      *  true #waitData# returns immediately.
      *  @memo Wait for data or a signal.
      *  @param test_only Return immediately if true.
      *  @return true if data are availble.
      */
    bool waitData(bool test_only=false) const;

#ifndef __CINT__
    /**  Find an Adc data structure with the given channel name and
      *  return a pointer to the object. If the object isn't found,
      *  a null pointer is returned.
      */
    const FrameCPP::FrAdcData* findAdc(const std::string& name) const;

    /**  Look for an Adc data structure with the given channel name 
      *  starting at the last found Adc. Return a pointer to the object. 
      *  If the object isn't found, a null pointer is returned.
      */
    const FrameCPP::FrAdcData* nextAdc(const std::string& name) const;

    /**  Find a Processed Data structure with the given channel name and
      *  return a pointer to the object. If the object isn't found,
      *  a null pointer is returned.
      */
    const FrameCPP::FrProcData* findProc(const std::string& name) const;

    /**  Look for a Processed Data structure with the given channel name 
      *  starting at the last found ProcData. Return a pointer to the object. 
      *  If the object isn't found, a null pointer is returned.
      */
    const FrameCPP::FrProcData* nextProc(const std::string& name) const;

    /**  Find a Simulated Data structure with the given channel name and
      *  return a pointer to the object. If the object isn't found,
      *  a null pointer is returned.
      */
    const FrameCPP::FrSimData* findSim(const std::string& name) const;

    /**  Look for a Simulated Data structure with the given channel name 
      *  starting at the last found SimData. Return a pointer to the object. 
      *  If the object isn't found, a null pointer is returned.
      */
    const FrameCPP::FrSimData* nextSim(const std::string& name) const;

    /**  Find a Static Data structure with the given channel name and
      *  valid at the specified time and version. If the versionis negative,
      *  the latest version number is returned. A valid FrStatDataRef is 
      *  returned if found. If the object isn't found, the FrStatDataRef is  
      *  a null reference (i.e. FrStatDataRef::null() returns true).
      *  @param name Name of static data to be found
      *  @param gps  Time of validity
      *  @param vsn  Version number
      */
  FrStatDataRef findStat(const std::string& name, const Time& gps, int vsn=-1);
#endif

  private:
    int closeFile(void);
    int closeReader(void);
    int deleteFrame(void);
    int flushFile(void);
    int openFile(const char* Source);
    int openReader(void);
    int readFrame(void);

#ifndef __CINT__
    const FrameCPP::FrAdcData*  readAdc(const std::string& name) const;
    const FrameCPP::FrProcData* readProc(const std::string& name) const;
    const FrameCPP::FrSimData*  readSim(const std::string& name) const;
#endif

  private:
    FileList mFile;

    /**  Debug mode flag.
      */
    int mDebug;

    /**  Time of current (Next) Frame.
      *  @memo Time of current (Next) Frame.
      */
    Time mT0;

    /** Frame count in this file
      */
    unsigned int mFramesInFile;

    /** Frame number
      */
    unsigned int mFrameNo;

    /** Frame number
      */
    enum {s_noTOC, s_TOC}  mTOCMode;

    /**  Current data source.
      *  @memo Current data source.
      */
    enum {s_none, s_online, s_file} mSource;

    /**  Pointer to input stream buffer object.
      *  @memo Pointer to input stream buffer object.
      */
#ifndef __CINT__
    std::streambuf* mSbuf;
#endif
    /**  Pointer to input stream buffer object.
      *  @memo Pointer to input stream buffer object.
      */
    std::istream* mStream;

#ifndef __CINT__
    /**  Pointer to current frame reader.
      *  @memo Pointer to current frame reader.
      */
    FrameCPP::IFrameStream* mReader;

    /**  Pointer to current frame. If no frame hase been read, mFrame is
      *  set to NULL.
      *  @memo Pointer to current frame.
      */
    FrameCPP::FrameH* mFrame;

    mutable AdcDataIterator  mAdcIter;
    mutable ProcDataIterator mProcIter;
    mutable SimDataIterator  mSimIter;

    FrameCPP::FrStatData::Query mStatData;
#endif

    /**  The Frame ID format string is set the first time getFrameID() is
      *  called, and is used to generate an output frame ID.
      */
    std::string mIDFormat;

    long mTotalFrames;
};

//======================================  Inline methods
#ifndef __CINT__
inline void
DaccIn::addFile(const char* name) {
    return mFile.addFile(name);
}

inline void
DaccIn::addFileList(const char* name) {
    return mFile.addList(name);
}

inline const char*
DaccIn::getFile(void) const {
    return mFile.getFront();
}

inline std::ostream& 
DaccIn::listFiles(std::ostream& out) const {
    return mFile.print(out);
}

inline FileList& 
DaccIn::refList(void) {
    return mFile;
}

inline const FileList& 
DaccIn::refList(void) const {
    return mFile;
}
#endif // !defined(__CINT__)

#endif // DACCIN_HH
