#ifndef READTREND_HH
#define READTREND_HH

#include "TSeries.hh"
#include "framedir.hh"
#include "Dacc.hh"
#include <string>
#include <vector>

class Time;

/**  ReadTrend class reads in one or more trends and returns requested
  *  statistical fields as TSeries. The trend files to be read in are
  *  found from a #framedir# directory built from all files matching the
  *  format #<directory>/<prefix>*<postfix>#. The directory, prefix and 
  *  postfix may be specified explicitly, or a monitor name may be 
  *  specified in which case, the file name fields are derived from the 
  *  monitor name and local defaults. The following statistical values 
  *  can be extracted from the trend frames:
  *  \begin{tabular}{|l|l|}
  *  Name  & Contents \\ \hline
  *  count & Number of points in average \\
  *  error & Error on the mean \\
  *  max   & Maximum value \\
  *  mean  & Mean value. \\
  *  min   & Minimum value \\
  *  sigma & Standard deviation \\ \hline
  *  \end{tabular}
  *  @author J. Zweizig
  *  @version 1.2; Last modified: September 10, 2003
  */
class ReadTrend {
public:
  /**  Enumerate trend types
    */
  enum TrendType {
    kSecond,
    kMinute,
    kNonStandard
  };

  /**  Enumerated data fields
    */
  enum typeID {
     kMean,
     kSigma,
     kMin,
     kMax,
     kError,
     kCount
  };

  /**  Vector of channel names or trend type names.
   */
  typedef std::vector<std::string> string_vect;

  /**  Vector of time series.
   */
  typedef std::vector<TSeries>     tseries_vect;

  /**  Vector of data field IDs
   */
  typedef std::vector<typeID>      type_vect;

public:
  /**  Construct a null trend reader.
    *  @memo Default constructor.
    */
  ReadTrend(void);

  /**  Construct a trend reader object, specifying the file prefix and 
    *  postfix.
    *  @memo Data constructor.
    *  @param dir    A name or RE specifying the directory(ies) containing 
    *                the minute trends.
    *  @param prefix Trend file name prefix.
    *  @param Type   Trend type
    */
  ReadTrend(const char* dir, const char* prefix=0, 
	    const char* postfix=0, TrendType Type=kMinute);

  /**  Get a Time series containing the maximum values for each trend 
    *  increments for the specified channel.
    *  @memo Get the maximum values from the specified trend.
    */
  int getMaxSeries(const char* chan, const Time& t0, Interval dT, 
		   TSeries* Max);

  /**  Get the mean value and optionally the error on the mean for a 
    *  specified channel and time interval.
    *  @memo Get ean and error time series.
    *  @param chan Channel name
    *  @param t0   Start time.
    *  @param dT   Time interval.
    *  @param Avg  Average value time series.
    *  @param Err  Error time series.
    */
  int getSeries(const char* chan, const Time& t0, Interval dT, 
		TSeries* Avg, TSeries* Err=0);

  /**  Get a time series of one or more statistical quantities.
    *  @memo get Statistics time series.
    *  @param chan  Vector of channel names.
    *  @param field Vector of field names.
    *  @param t0    Start time.
    *  @param dT    Time interval.
    *  @param Data  Vector of time series to receive data.
    */
  int getSeries(const string_vect& chan, const string_vect& field, 
		const Time& t0, Interval dT, tseries_vect& Data);


  /**  Get time series of one or more statistical quantities extracted from 
    *  channels in a trend frame. The number of Channel/Quantity pairs
    *  is arbitrary. The error codes are as reported by Dacc::fillData().
    *  @memo get Statistics time series.
    *  @param chan  Vector of channel names.
    *  @param field Vector of field IDs.
    *  @param t0    Start time.
    *  @param dT    Time interval.
    *  @param Data  Vector of time series to receive data.
    *  @return Error code
    */
  int getSeries(const string_vect& chan, const type_vect& field, 
		const Time& t0, Interval dT, tseries_vect& Data);

  /**  Specify file name search pattern for appropriate data. #setMonitor#
    *  sets the file names as follows:
    *  \begin{tabular}{|l|l|} \hline
    *  #<directory># & #$DMTRENDOUT# \\
    *  #<prefix>#    & #<site>-<monitor>_[TM]-# \\
    *  #<postfix>#   & #-<duration>.gwf# \\
    *  \end{tabular}
    *  @memo Set file name search pattern
    *  @param mon Monitor name.
    */
  void setMonitor(const char* mon);

  /**  Set the trend type.
    *  @memo Set the trend type.
    *  @param Typ Trend type code.
    */
  void setType(TrendType Typ);

  /**  Set the debug flag to print a lot of debugging information.
    */
  void setDebug(void);

  /**  Remove all file name entries from the frame directory.
    *  @memo expunge frame directory.
    */
  void expunge(void);

private:
  void buildDirectory(const char* name);
  void open(const Time& t0, Interval dT);
  void close(void);

private:
  std::string mDirec;
  std::string mPrefix;
  std::string mPostfix;
  std::string mCurDir;
  TrendType   mType;
  FrameDir    mFrameDir;
  Dacc        mIn;
  bool        mDebug;
};

#endif // !def(READTREND_HH)
